# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import os
from sample_utilities import get_authority, get_credential, get_client_modifications
from azure.appconfiguration.provider import load, SettingSelector

endpoint = os.environ.get("APPCONFIGURATION_ENDPOINT_STRING")
authority = get_authority(endpoint)
credential = get_credential(authority)
kwargs = get_client_modifications()

# Connecting to Azure App Configuration using AAD
config = load(endpoint=endpoint, credential=credential, **kwargs)

print(config["message"])

# Connecting to Azure App Configuration using AAD and trim key prefixes
trimmed = ["test."]
config = load(endpoint=endpoint, credential=credential, trim_prefixes=trimmed, **kwargs)

print(config["message"])

# Connection to Azure App Configuration using SettingSelector
selects = [SettingSelector(key_filter="message*")]
config = load(
    endpoint=endpoint,
    credential=credential,
    selects=selects,
    feature_flag_enabled=True,
    feature_flag_selectors=None,
    **kwargs
)

print("message found: " + str("message" in config))
print("test.message found: " + str("test.message" in config))
print("feature_flag_enabled found: " + str(config.get("feature_management")))



# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import unittest
from unittest.mock import patch, MagicMock
import datetime

from azure.core.exceptions import HttpResponseError

from azure.appconfiguration.provider._azureappconfigurationproviderbase import (
    _get_startup_backoff,
    _calculate_backoff_duration,
)
from azure.appconfiguration.provider._constants import (
    DEFAULT_STARTUP_TIMEOUT,
    MAX_STARTUP_BACKOFF_DURATION,
    MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION,
    STARTUP_BACKOFF_INTERVALS,
    JITTER_RATIO,
)


class TestGetFixedBackoff(unittest.TestCase):
    """Test the _get_startup_backoff function."""

    def test_within_first_interval(self):
        """Test backoff within first 100 seconds returns 5 seconds."""
        # STARTUP_BACKOFF_INTERVALS[0] = (100, 5)
        backoff, exceeded = _get_startup_backoff(0, 1)
        self.assertEqual(backoff, 5)
        self.assertFalse(exceeded)

        backoff, exceeded = _get_startup_backoff(50, 1)
        self.assertEqual(backoff, 5)
        self.assertFalse(exceeded)

        backoff, exceeded = _get_startup_backoff(99, 1)
        self.assertEqual(backoff, 5)
        self.assertFalse(exceeded)

    def test_within_second_interval(self):
        """Test backoff within 100-200 seconds returns 10 seconds."""
        # STARTUP_BACKOFF_INTERVALS[1] = (200, 10)
        backoff, exceeded = _get_startup_backoff(100, 1)
        self.assertEqual(backoff, 10)
        self.assertFalse(exceeded)

        backoff, exceeded = _get_startup_backoff(150, 1)
        self.assertEqual(backoff, 10)
        self.assertFalse(exceeded)

        backoff, exceeded = _get_startup_backoff(199, 1)
        self.assertEqual(backoff, 10)
        self.assertFalse(exceeded)

    def test_within_third_interval(self):
        """Test backoff within 200-600 seconds returns MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION."""
        # STARTUP_BACKOFF_INTERVALS[2] = (600, MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION)
        backoff, exceeded = _get_startup_backoff(200, 1)
        self.assertEqual(backoff, MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION)
        self.assertFalse(exceeded)

        backoff, exceeded = _get_startup_backoff(400, 1)
        self.assertEqual(backoff, MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION)
        self.assertFalse(exceeded)

        backoff, exceeded = _get_startup_backoff(599, 1)
        self.assertEqual(backoff, MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION)
        self.assertFalse(exceeded)

    def test_beyond_fixed_window_uses_exponential_backoff(self):
        """Test backoff beyond 600 seconds returns exponential backoff based on attempts."""
        # Beyond fixed window, uses _calculate_backoff_duration(attempts)
        # For attempt 1, _calculate_backoff_duration(1) does attempts+=1 -> 2, so jittered(30 * 2^1 = 60)
        backoff, exceeded = _get_startup_backoff(600, 1)
        self.assertTrue(exceeded)
        self.assertGreaterEqual(backoff, MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION * (1 - JITTER_RATIO))
        self.assertLessEqual(backoff, 60 * (1 + JITTER_RATIO))

        # For attempt 2, _calculate_backoff_duration(2) does attempts+=1 -> 3, so jittered(30 * 2^2 = 120)
        backoff, exceeded = _get_startup_backoff(1000, 2)
        self.assertTrue(exceeded)
        self.assertGreaterEqual(backoff, MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION * (1 - JITTER_RATIO))
        self.assertLessEqual(backoff, 120 * (1 + JITTER_RATIO))

    def test_negative_elapsed_time(self):
        """Test negative elapsed time returns first interval backoff."""
        backoff, exceeded = _get_startup_backoff(-1, 1)
        self.assertEqual(backoff, 5)
        self.assertFalse(exceeded)

    def test_boundary_conditions(self):
        """Test exact boundary values."""
        # Test at exact threshold boundaries
        for threshold, expected_backoff in STARTUP_BACKOFF_INTERVALS:
            # Just before threshold should return expected_backoff
            backoff, exceeded = _get_startup_backoff(threshold - 0.001, 1)
            self.assertEqual(backoff, expected_backoff)
            self.assertFalse(exceeded)


class TestCalculateBackoffDuration(unittest.TestCase):
    """Test the _calculate_backoff_duration function."""

    def test_first_attempt_returns_min_duration(self):
        """Test that first attempt returns minimum duration."""
        # attempts=0 -> internal attempts=1 after +=1, returns MIN
        result = _calculate_backoff_duration(0)
        self.assertEqual(result, MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION)

    def test_exponential_growth(self):
        """Test that backoff grows exponentially."""
        # Input attempts are shifted by +1 internally
        # For attempts=1, internal=2, calculated = 30 * 2^1 = 60
        result2 = _calculate_backoff_duration(1)
        # Result should be within jitter range of calculated value
        self.assertGreaterEqual(result2, MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION * (1 - JITTER_RATIO))
        self.assertLessEqual(result2, 60 * (1 + JITTER_RATIO))

        # For attempts=2, internal=3, calculated = 30 * 2^2 = 120
        result3 = _calculate_backoff_duration(2)
        self.assertGreaterEqual(result3, MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION * (1 - JITTER_RATIO))
        self.assertLessEqual(result3, 120 * (1 + JITTER_RATIO))

        # For attempts=3, internal=4, calculated = 30 * 2^3 = 240
        result4 = _calculate_backoff_duration(3)
        self.assertGreaterEqual(result4, MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION * (1 - JITTER_RATIO))
        self.assertLessEqual(result4, 240 * (1 + JITTER_RATIO))

    def test_caps_at_max_duration(self):
        """Test that backoff is capped at max duration."""
        # For attempts=5, internal=6, calculated = 30 * 2^5 = 960, but should cap at MAX_STARTUP_BACKOFF_DURATION (600)
        result = _calculate_backoff_duration(5)
        self.assertLessEqual(result, MAX_STARTUP_BACKOFF_DURATION * (1 + JITTER_RATIO))

    def test_invalid_attempts_raises_error(self):
        """Test that attempts < 1 raises ValueError."""
        # Input -1 -> internal 0, which is < 1
        with self.assertRaises(ValueError) as context:
            _calculate_backoff_duration(-1)
        self.assertIn("Number of attempts must be at least 1", str(context.exception))

        with self.assertRaises(ValueError):
            _calculate_backoff_duration(-2)

    def test_very_large_attempts_does_not_overflow(self):
        """Test that very large attempt numbers don't cause overflow."""
        # Should not raise an exception and should return a reasonable value
        result = _calculate_backoff_duration(100)
        self.assertGreater(result, 0)
        self.assertLessEqual(result, MAX_STARTUP_BACKOFF_DURATION * (1 + JITTER_RATIO))


class TestStartupRetryIntegration(unittest.TestCase):
    """Integration tests for startup retry behavior."""

    def test_default_startup_timeout_value(self):
        """Test that DEFAULT_STARTUP_TIMEOUT has expected value."""
        self.assertEqual(DEFAULT_STARTUP_TIMEOUT, 100)

    def test_max_backoff_duration_value(self):
        """Test that MAX_STARTUP_BACKOFF_DURATION has expected value."""
        self.assertEqual(MAX_STARTUP_BACKOFF_DURATION, 600)

    def test_MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION_value(self):
        """Test that MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION has expected value."""
        self.assertEqual(MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION, 30)

    def test_startup_backoff_intervals_structure(self):
        """Test that STARTUP_BACKOFF_INTERVALS has expected structure."""
        self.assertEqual(len(STARTUP_BACKOFF_INTERVALS), 3)
        # Should be sorted by threshold
        thresholds = [interval[0] for interval in STARTUP_BACKOFF_INTERVALS]
        self.assertEqual(thresholds, sorted(thresholds))

    def test_jitter_ratio_value(self):
        """Test that JITTER_RATIO has expected value."""
        self.assertEqual(JITTER_RATIO, 0.25)


class TestLoadAllRetryBehavior(unittest.TestCase):
    """Test the _load_all method retry behavior."""

    def setUp(self):
        """Set up test fixtures."""
        # We'll use mocking to test the retry behavior
        pass

    @patch("azure.appconfiguration.provider._azureappconfigurationprovider.time.sleep")
    @patch("azure.appconfiguration.provider._azureappconfigurationprovider.datetime")
    def test_load_all_timeout_raises_timeout_error(self, mock_datetime, mock_sleep):
        """Test that _load_all raises TimeoutError when startup_timeout is exceeded."""
        from azure.appconfiguration.provider._azureappconfigurationprovider import AzureAppConfigurationProvider
        from azure.appconfiguration.provider._azureappconfigurationproviderbase import AzureAppConfigurationProviderBase

        # Create a mock provider - use a custom __init__ that calls base but doesn't set up client manager
        def mock_init(self, **kwargs):
            AzureAppConfigurationProviderBase.__init__(self, endpoint="http://test.endpoint", **kwargs)
            self._startup_timeout = 1  # 1 second timeout
            self._replica_client_manager = MagicMock()
            self._secret_provider = MagicMock()
            self._secret_provider.uses_key_vault = False

        with patch.object(AzureAppConfigurationProvider, "__init__", mock_init):
            provider = AzureAppConfigurationProvider()
            provider._replica_client_manager.get_next_active_client.return_value = None

            # Mock datetime to simulate time passing
            start_time = datetime.datetime(2026, 1, 1, 0, 0, 0)
            times = [
                start_time,  # Initial startup_start_time
                start_time,  # First loop elapsed check (0 seconds elapsed)
                start_time,  # First loop remaining timeout check
                start_time + datetime.timedelta(seconds=2),  # Second loop check exceeds timeout
            ]
            mock_datetime.datetime.now.side_effect = times

            with self.assertRaises(TimeoutError) as context:
                provider._load_all()

            self.assertIn("timed out", str(context.exception.args[0]).lower())

    @patch("azure.appconfiguration.provider._azureappconfigurationprovider.time.sleep")
    def test_load_all_uses_fixed_backoff_initially(self, mock_sleep):
        """Test that _load_all uses fixed backoff during initial period."""
        from azure.appconfiguration.provider._azureappconfigurationprovider import AzureAppConfigurationProvider
        from azure.appconfiguration.provider._azureappconfigurationproviderbase import AzureAppConfigurationProviderBase

        # Create a mock provider - use a custom __init__ that calls base but doesn't set up client manager
        def mock_init(self, **kwargs):
            AzureAppConfigurationProviderBase.__init__(self, endpoint="http://test.endpoint", **kwargs)
            self._startup_timeout = 10
            self._replica_client_manager = MagicMock()
            self._secret_provider = MagicMock()
            self._secret_provider.uses_key_vault = False

        with patch.object(AzureAppConfigurationProvider, "__init__", mock_init):
            provider = AzureAppConfigurationProvider()

            # First attempt fails, second succeeds
            attempt_count = [0]

            def mock_try_initialize(startup_exceptions, **kwargs):
                attempt_count[0] += 1
                if attempt_count[0] == 1:
                    startup_exceptions.append(ConnectionError("Connection failed"))
                    return False
                return True

            provider._try_initialize = mock_try_initialize

            provider._load_all()

            # Verify sleep was called with fixed backoff duration (5 seconds for first 100 seconds)
            mock_sleep.assert_called_once()
            sleep_duration = mock_sleep.call_args[0][0]
            self.assertEqual(sleep_duration, 5)

    @patch("azure.appconfiguration.provider._azureappconfigurationprovider.time.sleep")
    def test_load_all_succeeds_on_first_try(self, mock_sleep):
        """Test that _load_all succeeds without retries when first attempt succeeds."""
        from azure.appconfiguration.provider._azureappconfigurationprovider import AzureAppConfigurationProvider
        from azure.appconfiguration.provider._azureappconfigurationproviderbase import AzureAppConfigurationProviderBase

        # Create a mock provider - use a custom __init__ that calls base but doesn't set up client manager
        def mock_init(self, **kwargs):
            AzureAppConfigurationProviderBase.__init__(self, endpoint="http://test.endpoint", **kwargs)
            self._startup_timeout = 100
            self._replica_client_manager = MagicMock()
            self._secret_provider = MagicMock()
            self._secret_provider.uses_key_vault = False

        with patch.object(AzureAppConfigurationProvider, "__init__", mock_init):
            provider = AzureAppConfigurationProvider()

            # First attempt succeeds
            provider._try_initialize = MagicMock(return_value=True)

            provider._load_all()

            # Sleep should not be called when first attempt succeeds
            mock_sleep.assert_not_called()
            provider._try_initialize.assert_called_once()


class TestStartupTimeoutConfiguration(unittest.TestCase):
    """Test startup_timeout configuration."""

    def test_custom_startup_timeout(self):
        """Test that custom startup_timeout is respected."""
        from azure.appconfiguration.provider._azureappconfigurationprovider import _buildprovider

        with patch("azure.appconfiguration.provider._azureappconfigurationprovider.ConfigurationClientManager"):
            with patch("azure.appconfiguration.provider._azureappconfigurationprovider.SecretProvider"):
                provider = _buildprovider(
                    connection_string="Endpoint=https://fake.endpoint;Id=fake;Secret=fake",
                    endpoint=None,
                    credential=None,
                    startup_timeout=60,
                )
                self.assertEqual(provider._startup_timeout, 60)

    def test_default_startup_timeout(self):
        """Test that default startup_timeout is used when not specified."""
        from azure.appconfiguration.provider._azureappconfigurationprovider import _buildprovider

        with patch("azure.appconfiguration.provider._azureappconfigurationprovider.ConfigurationClientManager"):
            with patch("azure.appconfiguration.provider._azureappconfigurationprovider.SecretProvider"):
                provider = _buildprovider(
                    connection_string="Endpoint=https://fake.endpoint;Id=fake;Secret=fake",
                    endpoint=None,
                    credential=None,
                )
                self.assertEqual(provider._startup_timeout, DEFAULT_STARTUP_TIMEOUT)


if __name__ == "__main__":
    unittest.main()



# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------

import time
from urllib.parse import urlparse
import dns.resolver
from dns.resolver import NXDOMAIN, YXDOMAIN, LifetimeTimeout, NoNameservers  # cspell:disable-line
from dns.rdatatype import SRV  # cspell:disable-line

request_retry_period = 5  # seconds
HTTPS_PREFIX = "https://"


class SRVRecord:
    priority: int
    weight: int
    port: int
    target: str

    def __init__(self, answer):
        self.priority = answer.priority
        self.weight = answer.weight
        self.port = answer.port
        self.target = str(answer.target).rstrip(".")


def find_auto_failover_endpoints(endpoint: str, replica_discovery_enabled: bool):
    if not replica_discovery_enabled:
        return []
    known_domain = _get_known_domain(endpoint)
    if known_domain is None:
        return []

    origin = _find_origin(endpoint)
    if origin is None or not _validate(known_domain, origin.target):
        return []

    replicas = _find_replicas(origin.target)

    if not replicas:
        return None  # Timeout

    srv_records = [origin] + replicas
    endpoints = []
    if endpoint.startswith(HTTPS_PREFIX):
        endpoint = endpoint[len(HTTPS_PREFIX) :]
    for srv_record in srv_records:
        if _validate(known_domain, srv_record.target) and srv_record.target != endpoint:
            endpoints.append(HTTPS_PREFIX + srv_record.target)
    return endpoints


def _find_origin(endpoint):
    uri = urlparse(endpoint).hostname
    request = f"_origin._tcp.{uri}"
    srv_records = _request_record(request)
    if not srv_records:
        return None
    return SRVRecord(srv_records[0])


def _find_replicas(origin):
    replicas = []
    i = 0
    while True:
        request = f"_alt{str(i)}._tcp.{origin}"
        answers = _request_record(request)
        if answers is None:
            return None  # Timeout
        if len(answers) == 0:
            break
        for answer in answers:
            replicas.append(SRVRecord(answer))
        i += 1
    return replicas


def _request_record(request):
    start_time = time.time()
    while time.time() - start_time < request_retry_period:
        try:
            return dns.resolver.resolve(request, SRV)
        except (NXDOMAIN, YXDOMAIN):  # cspell:disable-line
            return []  # No records found
        except (LifetimeTimeout, NoNameservers):
            continue
    return None  # Timeout


def _validate(known_domain, endpoint):
    return endpoint.endswith(known_domain)


def _get_known_domain(known_host):
    trusted_domain_labels = ["appconfig", "azconfig"]  # cspell:disable-line

    for label in trusted_domain_labels:
        index = known_host.lower().rfind(f".{label}.")
        if index > 0:
            return known_host[index:]
    return None



# coding: utf-8
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from devtools_testutils import AzureRecordedTestCase
from azure.appconfiguration.aio import AzureAppConfigurationClient
from testcase import get_configs
from azure.appconfiguration.provider.aio import load
from azure.appconfiguration.provider import AzureAppConfigurationKeyVaultOptions


class AppConfigTestCase(AzureRecordedTestCase):
    async def create_client(self, **kwargs):
        credential = self.get_credential(AzureAppConfigurationClient, is_async=True)
        client = None

        if "connection_string" in kwargs:
            client = AzureAppConfigurationClient.from_connection_string(kwargs["connection_string"])
        else:
            client = AzureAppConfigurationClient(kwargs["endpoint"], credential)

        await setup_configs(client, kwargs.get("keyvault_secret_url"), kwargs.get("keyvault_secret_url2"))
        kwargs["user_agent"] = "SDK/Integration"

        if "endpoint" in kwargs:
            kwargs["credential"] = credential

        if "secret_resolver" not in kwargs and kwargs.get("keyvault_secret_url") and "key_vault_options" not in kwargs:
            kwargs["keyvault_credential"] = credential

        if "key_vault_options" in kwargs:
            key_vault_options = kwargs.pop("key_vault_options")
            if not key_vault_options.secret_resolver:
                key_vault_options = AzureAppConfigurationKeyVaultOptions(credential=credential)
            kwargs["key_vault_options"] = key_vault_options

        return await load(**kwargs)

    @staticmethod
    def create_sdk_client(appconfiguration_connection_string):
        return AzureAppConfigurationClient.from_connection_string(
            appconfiguration_connection_string, user_agent="SDK/Integration"
        )

    def create_aad_sdk_client(self, appconfiguration_endpoint_string):
        cred = self.get_credential(AzureAppConfigurationClient, is_async=True)
        return AzureAppConfigurationClient(appconfiguration_endpoint_string, cred, user_agent="SDK/Integration")


async def setup_configs(client, keyvault_secret_url, keyvault_secret_url2):
    async with client:
        for config in get_configs(keyvault_secret_url, keyvault_secret_url2):
            await client.set_configuration_setting(config)


async def cleanup_test_resources_async(
    client,
    settings=None,
    snapshot_names=None,
):
    """
    Clean up test resources from Azure App Configuration (async version).

    :param client: The async AzureAppConfigurationClient to use for cleanup.
    :param settings: List of ConfigurationSetting objects to delete.
    :param snapshot_names: List of snapshot names to archive.
    """
    # Archive snapshots
    if snapshot_names:
        for snapshot_name in snapshot_names:
            try:
                await client.archive_snapshot(snapshot_name)
            except Exception:
                pass

    # Delete configuration settings and feature flags
    if settings:
        for setting in settings:
            try:
                await client.delete_configuration_setting(key=setting.key, label=setting.label)
            except Exception:
                pass


async def set_test_settings_async(client, settings):
    """
    Set multiple configuration settings in Azure App Configuration (async version).

    :param client: The async AzureAppConfigurationClient to use.
    :param settings: List of ConfigurationSetting or FeatureFlagConfigurationSetting objects to set.
    """
    for setting in settings:
        await client.set_configuration_setting(setting)


async def create_snapshot_async(client, snapshot_name, key_filters, composition_type=None, retention_period=3600):
    """
    Create a snapshot in Azure App Configuration and verify it was created successfully (async version).

    :param client: The async AzureAppConfigurationClient to use.
    :param snapshot_name: The name of the snapshot to create.
    :param key_filters: List of key filter strings to define what settings to include.
    :param composition_type: The composition type for the snapshot (default: SnapshotComposition.KEY).
    :param retention_period: The retention period in seconds (default: 3600, minimum valid value).
    :return: The created snapshot.
    """
    from azure.appconfiguration import SnapshotComposition, ConfigurationSettingsFilter, SnapshotStatus
    from devtools_testutils import is_live

    if composition_type is None:
        composition_type = SnapshotComposition.KEY

    # Convert key filter strings to ConfigurationSettingsFilter objects
    filters = [ConfigurationSettingsFilter(key=key_filter) for key_filter in key_filters]

    snapshot_poller = await client.begin_create_snapshot(
        name=snapshot_name,
        filters=filters,
        composition_type=composition_type,
        retention_period=retention_period,
    )
    snapshot = await snapshot_poller.result()

    # Verify snapshot was created successfully
    assert snapshot.status == SnapshotStatus.READY, f"Snapshot status is {snapshot.status}, expected READY"
    assert (
        snapshot.composition_type == composition_type
    ), f"Snapshot composition_type is {snapshot.composition_type}, expected {composition_type}"

    # Verify snapshot name (sanitized in playback mode)
    if is_live():
        assert snapshot.name == snapshot_name, f"Snapshot name is {snapshot.name}, expected {snapshot_name}"
    else:
        assert snapshot.name == "Sanitized", f"Snapshot name is {snapshot.name}, expected 'Sanitized' in playback"

    return snapshot



# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import os
import time
from sample_utilities import get_client_modifications
from azure.appconfiguration import (  # type:ignore
    AzureAppConfigurationClient,
    ConfigurationSetting,
)
from sample_utilities import get_client_modifications
import os
import time
import random
from azure.appconfiguration.provider import load, WatchKey

kwargs = get_client_modifications()
connection_string = os.environ.get("APPCONFIGURATION_CONNECTION_STRING")

# Setting up a configuration setting with a known value
client = AzureAppConfigurationClient.from_connection_string(connection_string)

configuration_setting = ConfigurationSetting(key="message", value="Hello World!")

client.set_configuration_setting(configuration_setting=configuration_setting)


def my_callback_on_fail(_):
    print("Refresh failed!")


rand = random.random()
watch_key = WatchKey("message" + str(rand))

# Connecting to Azure App Configuration using connection string, and refreshing when the configuration setting message
# changes
config = load(
    connection_string=connection_string,
    refresh_on=[watch_key],
    refresh_interval=1,
    on_refresh_error=my_callback_on_fail,
    **kwargs,
)

print(config["message"])
print(config["my_json"]["key"])

# Updating the configuration setting
configuration_setting.value = "Hello World Updated!"

configuration_setting2 = ConfigurationSetting(key="message" + str(rand), value="2")

client.set_configuration_setting(configuration_setting=configuration_setting2)

client.set_configuration_setting(configuration_setting=configuration_setting)

# Waiting for the refresh interval to pass
time.sleep(2)

# Refreshing the configuration setting
config.refresh()

# Printing the updated value
print(config["message"])
print(config["my_json"]["key"])

# Waiting for the refresh interval to pass
time.sleep(2)

# Refreshing the configuration setting with no changes
config.refresh()

# Printing the updated value
print(config["message"])
print(config["my_json"]["key"])



# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------


# ------------------------------------------------------------------------
# Feature Management Constants
# ------------------------------------------------------------------------
FEATURE_MANAGEMENT_KEY = "feature_management"
FEATURE_FLAG_KEY = "feature_flags"
FEATURE_FLAG_PREFIX = ".appconfig.featureflag/"
FEATURE_FLAG_REFERENCE_KEY = "FeatureFlagReference"
ALLOCATION_ID_KEY = "AllocationId"
ETAG_KEY = "ETag"

# ------------------------------------------------------------------------
# Environment Variable Constants
# ------------------------------------------------------------------------
REQUEST_TRACING_DISABLED_ENVIRONMENT_VARIABLE = "AZURE_APP_CONFIGURATION_TRACING_DISABLED"
AzureFunctionEnvironmentVariable = "FUNCTIONS_EXTENSION_VERSION"
AzureWebAppEnvironmentVariable = "WEBSITE_SITE_NAME"
ContainerAppEnvironmentVariable = "CONTAINER_APP_NAME"
KubernetesEnvironmentVariable = "KUBERNETES_PORT"
ServiceFabricEnvironmentVariable = "Fabric_NodeName"  # cspell:disable-line

# ------------------------------------------------------------------------
# Telemetry and Tracing Constants
# ------------------------------------------------------------------------
TELEMETRY_KEY = "telemetry"
METADATA_KEY = "metadata"
SNAPSHOT_REFERENCE_TAG = "SnapshotRef"

# ------------------------------------------------------------------------
# Content Type and Mime Profile Constants
# ------------------------------------------------------------------------
APP_CONFIG_AI_MIME_PROFILE = "https://azconfig.io/mime-profiles/ai/"
APP_CONFIG_AICC_MIME_PROFILE = "https://azconfig.io/mime-profiles/ai/chat-completion"

# =============================================================================
# Startup Retry Constants
# =============================================================================
# Timeout
DEFAULT_STARTUP_TIMEOUT = 100  # seconds

# Backoff durations
MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION = 30  # seconds - minimum backoff after fixed window
MAX_STARTUP_BACKOFF_DURATION = 600  # seconds (10 minutes) - caps exponential backoff

# Jitter ratio for randomizing backoff durations (+/- 25% variation)
JITTER_RATIO = 0.25

# Fixed backoff intervals: (elapsed_time_threshold, backoff_duration)
# Defines fixed backoff durations based on how long startup has been attempting
STARTUP_BACKOFF_INTERVALS = [
    (100, 5),  # 0-100 seconds elapsed: 5 second backoff
    (200, 10),  # 100-200 seconds elapsed: 10 second backoff
    (600, MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION),  # 200-600 seconds elapsed: 30 second backoff
]

# ------------------------------------------------------------------------
# Snapshot Reference Constants
# ------------------------------------------------------------------------
SNAPSHOT_REF_CONTENT_TYPE = 'application/json; profile="https://azconfig.io/mime-profiles/snapshot-ref"; charset=utf-8'
SNAPSHOT_NAME_FIELD = "snapshot_name"

# ------------------------------------------------------------------------
# Miscellaneous Constants
# ------------------------------------------------------------------------
NULL_CHAR = "\0"



import os
from devtools_testutils import (
    add_general_regex_sanitizer,
    add_general_string_sanitizer,
    add_oauth_response_sanitizer,
    set_custom_default_matcher,
    remove_batch_sanitizers,
    add_remove_header_sanitizer,
    add_uri_string_sanitizer,
)
import pytest

# autouse=True will trigger this fixture on each pytest run, even if it's not explicitly used by a test method


@pytest.fixture(scope="session", autouse=True)
def add_sanitizers(test_proxy):
    add_general_regex_sanitizer(
        value="https://Sanitized.azconfig.io",
        regex=os.environ.get("APPCONFIGURATION_ENDPOINT_STRING", "https://Sanitized.azconfig.io"),
    )
    add_general_regex_sanitizer(
        value="Sanitized",
        regex=os.environ.get("APPCONFIGURATION_CONNECTION_STRING", "Sanitized"),
    )
    add_uri_string_sanitizer()
    add_general_string_sanitizer(
        value="https://Sanitized.vault.azure.net/secrets/fake-secret/",
        target=os.environ.get(
            "APPCONFIGURATION_KEY_VAULT_REFERENCE", "https://Sanitized.vault.azure.net/secrets/fake-secret/"
        ),
    )
    add_remove_header_sanitizer(headers="Correlation-Context")

    add_general_regex_sanitizer(value="api-version=1970-01-01", regex="api-version=.+")
    set_custom_default_matcher(ignored_headers="x-ms-content-sha256, Accept", excluded_headers="Content-Length")
    add_remove_header_sanitizer(headers="Sync-Token")
    add_oauth_response_sanitizer()

    # Remove the following sanitizers since certain fields are needed in tests and are non-sensitive:
    #  - AZSDK3430: $..id
    #  - AZSDK3447: $.key
    remove_batch_sanitizers(["AZSDK3430", "AZSDK3447"])



# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from typing import Optional, Callable, TYPE_CHECKING, Union, Awaitable, Mapping, Any, NamedTuple, List
from ._constants import NULL_CHAR

# Error message constants for SettingSelector validation
_SNAPSHOT_PARAMETER_CONFLICT_ERROR = (
    "Cannot specify both snapshot_name and {parameter}. "
    "When using snapshots, all other filtering parameters are ignored."
)
_SELECTOR_REQUIRED_ERROR = "Either key_filter or snapshot_name must be specified."

if TYPE_CHECKING:
    from azure.core.credentials import TokenCredential
    from azure.core.credentials_async import AsyncTokenCredential


class AzureAppConfigurationKeyVaultOptions:
    def __init__(
        self,
        *,
        credential: Optional[Union["TokenCredential", "AsyncTokenCredential"]] = None,
        client_configs: Optional[Mapping[str, Mapping[str, Any]]] = None,
        secret_resolver: Optional[Union[Callable[[str], str], Callable[[str], Awaitable[str]]]] = None,
    ):
        """
        Options for connecting to Key Vault.

        :keyword credential: A credential for authenticating with the key vault. This is optional if secret_clients is
         provided.
        :paramtype credential: ~azure.core.credentials.TokenCredential
        :keyword client_configs: A Mapping of SecretClient endpoints to client configurations from
         azure-keyvault-secrets. This is optional if credential is provided. If a credential isn't provided a
         credential will need to be in each set for each.
        :paramtype client_configs: Mapping[Url, Mapping]
        :keyword secret_resolver: A function that takes a URI and returns a value.
        :paramtype secret_resolver: Callable[[str], str]
        """
        self.credential = credential
        self.client_configs = client_configs or {}
        self.secret_resolver = secret_resolver
        if self.credential is not None and self.secret_resolver is not None:
            raise ValueError("credential and secret_resolver can't both be configured.")


class SettingSelector:
    """
    Selects a set of configuration settings from Azure App Configuration.

    :keyword key_filter: A filter to select configuration settings and feature flags based on their keys.
     Cannot be used with snapshot_name.
    :type key_filter: str
    :keyword label_filter: A filter to select configuration settings and feature flags based on their labels. Default
     value is \0 i.e. (No Label) as seen in the portal. Cannot be used with snapshot_name.
    :type label_filter: Optional[str]
    :keyword tag_filters: A filter to select configuration settings and feature flags based on their tags. This is a
     list of strings that will be used to match tags on the configuration settings. Reserved characters (\\*, \\, ,)
     must be escaped with backslash if they are part of the value. Tag filters must follow the format
     "tagName=tagValue", for empty values use "tagName=" and for null values use "tagName=\\0".
     Cannot be used with snapshot_name.
    :type tag_filters: Optional[List[str]]
    :keyword snapshot_name: The name of the snapshot to load configuration settings from. When specified,
     all configuration settings from the snapshot will be loaded. Cannot be used with key_filter, label_filter,
     or tag_filters.
    :type snapshot_name: Optional[str]
    """

    def __init__(
        self,
        *,
        key_filter: Optional[str] = None,
        label_filter: Optional[str] = NULL_CHAR,
        tag_filters: Optional[List[str]] = None,
        snapshot_name: Optional[str] = None,
    ):
        if snapshot_name is not None:
            # When using snapshots, no other filtering parameters should be specified
            if key_filter is not None:
                raise ValueError(_SNAPSHOT_PARAMETER_CONFLICT_ERROR.format(parameter="key_filter"))
            if label_filter != NULL_CHAR:
                raise ValueError(_SNAPSHOT_PARAMETER_CONFLICT_ERROR.format(parameter="label_filter"))
            if tag_filters is not None:
                raise ValueError(_SNAPSHOT_PARAMETER_CONFLICT_ERROR.format(parameter="tag_filters"))

        if snapshot_name is None and key_filter is None:
            raise ValueError(_SELECTOR_REQUIRED_ERROR)

        if tag_filters is not None:
            if not isinstance(tag_filters, list):
                raise TypeError("tag_filters must be a list of strings.")
            for tag in tag_filters:
                if not tag:
                    raise ValueError("Tag filter cannot be an empty string or None.")
                if not isinstance(tag, str) or "=" not in tag or tag.startswith("="):
                    raise ValueError("Tag filter " + tag + ' does not follow the format "tagName=tagValue".')

        self.key_filter = key_filter
        self.label_filter = label_filter
        self.tag_filters = tag_filters
        self.snapshot_name = snapshot_name


class WatchKey(NamedTuple):
    key: str
    label: str = NULL_CHAR



# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import base64
import hashlib
import json
import time
import datetime
import random
from threading import Lock
import logging
from typing import (
    Any,
    Dict,
    Mapping,
    Optional,
    overload,
    List,
    Tuple,
    Union,
    Iterator,
    KeysView,
    ItemsView,
    ValuesView,
    TypeVar,
)
from azure.appconfiguration import (  # type:ignore # pylint:disable=no-name-in-module
    ConfigurationSetting,
    FeatureFlagConfigurationSetting,
)
from ._models import SettingSelector
from ._constants import (
    NULL_CHAR,
    TELEMETRY_KEY,
    METADATA_KEY,
    ETAG_KEY,
    FEATURE_FLAG_REFERENCE_KEY,
    ALLOCATION_ID_KEY,
    APP_CONFIG_AI_MIME_PROFILE,
    APP_CONFIG_AICC_MIME_PROFILE,
    FEATURE_MANAGEMENT_KEY,
    FEATURE_FLAG_KEY,
    DEFAULT_STARTUP_TIMEOUT,
    MAX_STARTUP_BACKOFF_DURATION,
    MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION,
    JITTER_RATIO,
    STARTUP_BACKOFF_INTERVALS,
)
from ._refresh_timer import _RefreshTimer
from ._request_tracing_context import _RequestTracingContext


JSON = Mapping[str, Any]
_T = TypeVar("_T")
logger = logging.getLogger(__name__)

min_uptime = 5


def delay_failure(start_time: datetime.datetime) -> None:
    """
    We want to make sure we are up a minimum amount of time before we kill the process.
    Otherwise, we could get stuck in a quick restart loop.

    :param start_time: The time when the process started.
    :type start_time: datetime.datetime
    """
    min_time = datetime.timedelta(seconds=min_uptime)
    current_time = datetime.datetime.now()
    if current_time - start_time < min_time:
        time.sleep((min_time - (current_time - start_time)).total_seconds())


def process_load_parameters(*args, **kwargs: Any) -> Dict[str, Any]:
    """
    Process and validate all load function parameters in one place.
    This consolidates the most obviously duplicated logic from both sync and async load functions.

    :param args: Positional arguments, either endpoint and credential, or connection string.
    :type args: Any
    :return: Dictionary containing processed parameters
    :rtype: Dict[str, Any]
    """
    endpoint: Optional[str] = kwargs.pop("endpoint", None)
    credential = kwargs.pop("credential", None)
    connection_string: Optional[str] = kwargs.pop("connection_string", None)
    start_time = datetime.datetime.now()

    # Handle positional arguments
    if len(args) > 2:
        raise TypeError(
            "Unexpected positional parameters. Please pass either endpoint and credential, or a connection string."
        )
    if len(args) == 1:
        if endpoint is not None:
            raise TypeError("Received multiple values for parameter 'endpoint'.")
        endpoint = args[0]
    elif len(args) == 2:
        if credential is not None:
            raise TypeError("Received multiple values for parameter 'credential'.")
        endpoint, credential = args

    # Validate endpoint/credential vs connection_string
    if (endpoint or credential) and connection_string:
        raise ValueError("Please pass either endpoint and credential, or a connection string.")

    # Process Key Vault options in one place
    key_vault_options = kwargs.pop("key_vault_options", None)
    if key_vault_options:
        if "keyvault_credential" in kwargs or "secret_resolver" in kwargs or "keyvault_client_configs" in kwargs:
            raise ValueError(
                "Key Vault configurations should only be set by either the key_vault_options or kwargs not both."
            )
        kwargs["keyvault_credential"] = key_vault_options.credential
        kwargs["secret_resolver"] = key_vault_options.secret_resolver
        kwargs["keyvault_client_configs"] = key_vault_options.client_configs

    if kwargs.get("keyvault_credential") is not None and kwargs.get("secret_resolver") is not None:
        raise ValueError("A keyvault credential and secret resolver can't both be configured.")

    # Determine Key Vault usage
    uses_key_vault = (
        "keyvault_credential" in kwargs
        or "keyvault_client_configs" in kwargs
        or "secret_resolver" in kwargs
        or kwargs.get("uses_key_vault", False)
    )

    # Get startup timeout
    startup_timeout = kwargs.pop("startup_timeout", DEFAULT_STARTUP_TIMEOUT)
    if startup_timeout < 0:
        raise ValueError("Startup timeout must be greater than or equal to 0 seconds.")

    return {
        "endpoint": endpoint,
        "credential": credential,
        "connection_string": connection_string,
        "uses_key_vault": uses_key_vault,
        "start_time": start_time,
        "startup_timeout": startup_timeout,
        "kwargs": kwargs,
    }


def is_json_content_type(content_type: str) -> bool:
    if not content_type:
        return False

    content_type = content_type.strip().lower()
    mime_type = content_type.split(";")[0].strip()

    type_parts = mime_type.split("/")
    if len(type_parts) != 2:
        return False

    (main_type, sub_type) = type_parts
    if main_type != "application":
        return False

    sub_types = sub_type.split("+")
    if "json" in sub_types:
        return True

    return False


def _build_watched_setting(setting: Union[str, Tuple[str, str]]) -> Tuple[str, str]:
    try:
        key, label = setting  # type:ignore
    except (IndexError, ValueError):
        key = str(setting)  # Ensure key is a string
        label = NULL_CHAR
    if "*" in key or "*" in label:
        raise ValueError("Wildcard key or label filters are not supported for refresh.")
    return key, label


def sdk_allowed_kwargs(kwargs):
    allowed_kwargs = [
        "audience",
        "headers",
        "request_id",
        "user_agent",
        "logging_enable",
        "logger",
        "response_encoding",
        "raw_request_hook",
        "raw_response_hook",
        "network_span_namer",
        "tracing_attributes",
        "permit_redirects",
        "redirect_max",
        "retry_total",
        "retry_connect",
        "retry_read",
        "retry_status",
        "retry_backoff_factor",
        "retry_backoff_max",
        "retry_mode",
        "timeout",
        "connection_timeout",
        "read_timeout",
        "connection_verify",
        "connection_cert",
        "proxies",
        "cookies",
        "connection_data_block_size",
    ]
    return {k: v for k, v in kwargs.items() if k in allowed_kwargs}


def _jitter(duration: float, ratio: float = JITTER_RATIO) -> float:
    """
    Apply jitter to a duration value.

    :param duration: The base duration in seconds.
    :type duration: float
    :param ratio: The jitter ratio (0 to 1). Default is 0.25 (25% jitter means +/- 25% variation).
    :type ratio: float
    :return: The jittered duration in seconds.
    :rtype: float
    """
    if ratio < 0 or ratio > 1:
        raise ValueError("Jitter ratio must be between 0 and 1.")
    if ratio == 0:
        return duration
    jitter = ratio * (random.random() * 2 - 1)
    return duration * (1 + jitter)


def _get_startup_backoff(elapsed_seconds: float, attempts: int) -> Tuple[float, bool]:
    """
    Get a backoff duration based on elapsed startup time.

    :param elapsed_seconds: The time elapsed since startup began, in seconds.
    :type elapsed_seconds: float
    :param attempts: The number of retry attempts made (1-based).
    :type attempts: int
    :return: A tuple where the first element is the backoff duration in seconds,
             and the second element indicates if the fixed backoff window has been exceeded.
    :rtype: Tuple[float, bool]
    """
    for threshold, backoff in STARTUP_BACKOFF_INTERVALS:
        if elapsed_seconds < threshold:
            return backoff, False
    return _calculate_backoff_duration(attempts), True


def _calculate_backoff_duration(attempts: int) -> float:
    """
    Calculate the jittered exponential backoff duration.

    :param attempts: The number of retry attempts made (1-based).
    :type attempts: int
    :return: The calculated backoff duration with jitter applied.
    :rtype: float
    """
    attempts += 1
    if attempts < 1:
        raise ValueError("Number of attempts must be at least 1.")

    if attempts == 1:
        return MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION

    # Calculate exponential backoff: min * 2^(attempts-1)
    # Cap the shift amount to prevent overflow
    safe_shift = min(attempts - 1, 63)
    calculated = MIN_STARTUP_EXPONENTIAL_BACKOFF_DURATION * (1 << safe_shift)

    # Cap at max duration
    if calculated > MAX_STARTUP_BACKOFF_DURATION or calculated <= 0:  # Check for overflow
        calculated = MAX_STARTUP_BACKOFF_DURATION

    return _jitter(calculated, JITTER_RATIO)


class AzureAppConfigurationProviderBase(Mapping[str, Union[str, JSON]]):  # pylint: disable=too-many-instance-attributes
    """
    Provides a dictionary-like interface to Azure App Configuration settings. Enables loading of sets of configuration
    settings from Azure App Configuration into a Python application. Enables trimming of prefixes from configuration
    keys. Enables resolution of Key Vault references in configuration settings.
    """

    def __init__(self, **kwargs: Any) -> None:
        self._origin_endpoint: str = kwargs.get("endpoint", "")
        self._dict: Dict[str, Any] = {}
        self._selects: List[SettingSelector] = kwargs.pop(
            "selects", [SettingSelector(key_filter="*", label_filter=NULL_CHAR)]
        )

        trim_prefixes: List[str] = kwargs.pop("trim_prefixes", [])
        self._trim_prefixes: List[str] = sorted(trim_prefixes, key=len, reverse=True)
        refresh_on: List[Tuple[str, str]] = kwargs.pop("refresh_on", None) or []
        self._watched_settings: Dict[Tuple[str, str], Optional[str]] = {
            _build_watched_setting(s): None for s in refresh_on
        }
        self._refresh_timer: _RefreshTimer = _RefreshTimer(**kwargs)
        self._feature_flag_enabled = kwargs.pop("feature_flag_enabled", False)
        self._feature_flag_selectors = kwargs.pop("feature_flag_selectors", None)
        if self._feature_flag_selectors is None:
            self._feature_flag_selectors = [SettingSelector(key_filter="*")]
        self._watched_feature_flags: Dict[Tuple[str, str], Optional[str]] = {}
        self._feature_flag_refresh_timer: _RefreshTimer = _RefreshTimer(**kwargs)
        self._feature_flag_refresh_enabled = kwargs.pop("feature_flag_refresh_enabled", False)
        self._tracing_context = _RequestTracingContext(kwargs.pop("load_balancing_enabled", False))
        self._update_lock = Lock()
        self._refresh_lock = Lock()

    def _process_key_name(self, config):
        trimmed_key = config.key
        # Trim the key if it starts with one of the prefixes provided
        for trim in self._trim_prefixes:
            if config.key.startswith(trim):
                trimmed_key = config.key[len(trim) :]
                break
        return trimmed_key

    def _update_ff_telemetry_metadata(
        self, endpoint: str, feature_flag: FeatureFlagConfigurationSetting, feature_flag_value: Dict
    ):
        """
        Add telemetry metadata to feature flag values.

        :param endpoint: The App Configuration endpoint URL.
        :type endpoint: str
        :param feature_flag: The feature flag configuration setting.
        :type feature_flag: FeatureFlagConfigurationSetting
        :param feature_flag_value: The feature flag value dictionary to update.
        :type feature_flag_value: Dict[str, Any]
        """
        if TELEMETRY_KEY not in feature_flag_value:
            # Initialize telemetry dictionary if not present
            feature_flag_value[TELEMETRY_KEY] = {}

        # Update telemetry metadata for application insights/logging in feature management
        if METADATA_KEY not in feature_flag_value[TELEMETRY_KEY]:
            feature_flag_value[TELEMETRY_KEY][METADATA_KEY] = {}
        feature_flag_value[TELEMETRY_KEY][METADATA_KEY][ETAG_KEY] = feature_flag.etag

        if feature_flag_value[TELEMETRY_KEY].get("enabled"):
            self._tracing_context.uses_telemetry = True
            if not endpoint.endswith("/"):
                endpoint += "/"
            feature_flag_reference = f"{endpoint}kv/{feature_flag.key}"
            if feature_flag.label and not feature_flag.label.isspace():
                feature_flag_reference += f"?label={feature_flag.label}"

            feature_flag_value[TELEMETRY_KEY][METADATA_KEY][FEATURE_FLAG_REFERENCE_KEY] = feature_flag_reference
            allocation_id = self._generate_allocation_id(feature_flag_value)
            if allocation_id:
                feature_flag_value[TELEMETRY_KEY][METADATA_KEY][ALLOCATION_ID_KEY] = allocation_id

        allocation = feature_flag_value.get("allocation")
        if allocation and allocation.get("seed"):
            self._tracing_context.uses_seed = True

        variants = feature_flag_value.get("variants")
        if variants:
            # Update Usage Data for Telemetry
            self._tracing_context.update_max_variants(len(variants))

    @staticmethod
    def _generate_allocation_id(feature_flag_value: Dict[str, JSON]) -> Optional[str]:
        """
        Generates an allocation ID for the specified feature.
        seed=123abc\ndefault_when_enabled=Control\npercentiles=0,Control,20;20,Test,100\nvariants=Control,standard;Test,special # pylint:disable=line-too-long

        :param  Dict[str, JSON] feature_flag_value: The feature to generate an allocation ID for.
        :rtype: str
        :return: The allocation ID.
        """

        allocation_id = ""
        allocated_variants = []

        allocation: Optional[JSON] = feature_flag_value.get("allocation")

        if not allocation:
            return None

        # Seed
        allocation_id = f"seed={allocation.get('seed', '')}"

        # DefaultWhenEnabled
        if "default_when_enabled" in allocation:
            allocated_variants.append(allocation.get("default_when_enabled"))

        allocation_id += f"\ndefault_when_enabled={allocation.get('default_when_enabled', '')}"

        # Percentile
        allocation_id += "\npercentiles="

        percentile = allocation.get("percentile")

        if percentile:
            percentile_allocations = sorted(
                (x for x in percentile if x.get("from") != x.get("to")),
                key=lambda x: x.get("from"),
            )

            for percentile_allocation in percentile_allocations:
                if "variant" in percentile_allocation:
                    allocated_variants.append(percentile_allocation.get("variant"))

            allocation_id += ";".join(
                f"{pa.get('from')}," f"{base64.b64encode(pa.get('variant').encode()).decode()}," f"{pa.get('to')}"
                for pa in percentile_allocations
            )

        if not allocated_variants and not allocation.get("seed"):
            return None

        # Variants
        allocation_id += "\nvariants="

        variants_value = feature_flag_value.get("variants")
        if variants_value and (isinstance(variants_value, list) or all(isinstance(v, dict) for v in variants_value)):
            if (
                allocated_variants
                and isinstance(variants_value, list)
                and all(isinstance(v, dict) for v in variants_value)
            ):
                sorted_variants: List[Dict[str, Any]] = sorted(
                    (v for v in variants_value if v.get("name") in allocated_variants),
                    key=lambda v: v.get("name"),
                )

                for v in sorted_variants:
                    allocation_id += f"{base64.b64encode(v.get('name', '').encode()).decode()},"
                    if "configuration_value" in v:
                        allocation_id += (
                            f"{json.dumps(v.get('configuration_value', ''), separators=(',', ':'), sort_keys=True)}"
                        )
                    allocation_id += ";"
                if sorted_variants:
                    allocation_id = allocation_id[:-1]

        # Create a sha256 hash of the allocation_id
        hash_object = hashlib.sha256(allocation_id.encode())
        hash_digest = hash_object.digest()

        # Encode the first 15 bytes in base64 url
        return base64.urlsafe_b64encode(hash_digest[:15]).decode()

    def __getitem__(self, key: str) -> Any:
        # pylint:disable=docstring-missing-param,docstring-missing-return,docstring-missing-rtype
        """
        Returns the value of the specified key.
        """
        with self._update_lock:
            return self._dict[key]

    def __iter__(self) -> Iterator[str]:
        return self._dict.__iter__()

    def __len__(self) -> int:
        return len(self._dict)

    def __contains__(self, __x: object) -> bool:
        # pylint:disable=docstring-missing-param,docstring-missing-return,docstring-missing-rtype
        """
        Returns True if the configuration settings contains the specified key.
        """
        return self._dict.__contains__(__x)

    def keys(self) -> KeysView[str]:
        """
        Returns a list of keys loaded from Azure App Configuration.

        :return: A list of keys loaded from Azure App Configuration.
        :rtype: KeysView[str]
        """
        with self._update_lock:
            return self._dict.keys()

    def items(self) -> ItemsView[str, Union[str, Mapping[str, Any]]]:
        """
        Returns a set-like object of key-value pairs loaded from Azure App Configuration. Any values that are Key Vault
         references will be resolved.

        :return: A set-like object of key-value pairs loaded from Azure App Configuration.
        :rtype: ItemsView[str, Union[str, Mapping[str, Any]]]
        """
        with self._update_lock:
            return self._dict.items()

    def values(self) -> ValuesView[Union[str, Mapping[str, Any]]]:
        """
        Returns a list of values loaded from Azure App Configuration. Any values that are Key Vault references will be
        resolved.

        :return: A list of values loaded from Azure App Configuration. The values are either Strings or JSON objects,
         based on there content type.
        :rtype: ValuesView[Union[str, Mapping[str, Any]]]
        """
        with self._update_lock:
            return (self._dict).values()

    @overload
    def get(self, key: str, default: None = None) -> Union[str, JSON, None]: ...

    @overload
    def get(self, key: str, default: Union[str, JSON, _T]) -> Union[str, JSON, _T]:  # pylint: disable=signature-differs
        ...

    def get(self, key: str, default: Optional[Union[str, JSON, _T]] = None) -> Union[str, JSON, _T, None]:
        """
        Returns the value of the specified key. If the key does not exist, returns the default value.

        :param key: The key of the value to get.
        :type key: str
        :param default: The default value to return.
        :type default: Optional[Union[str, JSON, _T]]
        :return: The value of the specified key.
        :rtype: Union[str, JSON, _T, None]
        """
        with self._update_lock:
            return self._dict.get(key, default)

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def _process_key_value_base(self, config: ConfigurationSetting) -> Union[str, Dict[str, Any]]:
        """
        Process configuration values that are not KeyVault references. If the content type is None, the value is
        returned as-is.

        :param config: The configuration setting to process.
        :type config: ConfigurationSetting
        :return: The processed configuration value (JSON object if JSON content type, string otherwise).
        :rtype: Union[str, Dict[str, Any]]
        """
        if config.content_type is None:
            return config.value
        if is_json_content_type(config.content_type) and not isinstance(config, FeatureFlagConfigurationSetting):
            # Feature flags are of type json, but don't treat them as such
            try:
                if APP_CONFIG_AI_MIME_PROFILE in config.content_type:
                    self._tracing_context.uses_ai_configuration = True
                if APP_CONFIG_AICC_MIME_PROFILE in config.content_type:
                    self._tracing_context.uses_aicc_configuration = True
                return json.loads(config.value)
            except json.JSONDecodeError:
                try:
                    # If the value is not a valid JSON, check if it has comments and remove them
                    from ._json import remove_json_comments

                    return json.loads(remove_json_comments(config.value))
                except (json.JSONDecodeError, ValueError):
                    # If the value is not a valid JSON, treat it like regular string value
                    return config.value
        return config.value

    def _process_feature_flags(
        self,
        processed_settings: Dict[str, Any],
        processed_feature_flags: List[Dict[str, Any]],
        feature_flags: Optional[List[FeatureFlagConfigurationSetting]],
    ) -> Dict[str, Any]:
        if feature_flags:
            # Reset feature flag usage
            self._tracing_context.reset_feature_filter_usage()
            processed_feature_flags = [self._process_feature_flag(ff) for ff in feature_flags]
            self._watched_feature_flags = self._update_watched_feature_flags(feature_flags)

        if self._feature_flag_enabled:
            processed_settings[FEATURE_MANAGEMENT_KEY] = {}
            processed_settings[FEATURE_MANAGEMENT_KEY][FEATURE_FLAG_KEY] = processed_feature_flags
        return processed_settings

    def _process_feature_flag(self, feature_flag: FeatureFlagConfigurationSetting) -> Dict[str, Any]:
        try:
            feature_flag_value = json.loads(feature_flag.value)
            self._update_ff_telemetry_metadata(self._origin_endpoint, feature_flag, feature_flag_value)
            self._tracing_context.update_feature_filter_telemetry(feature_flag)
            return feature_flag_value
        except json.JSONDecodeError:
            # Feature flag value is not a valid JSON
            return {}

    def _update_watched_settings(
        self, configuration_settings: List[ConfigurationSetting]
    ) -> Dict[Tuple[str, str], Optional[str]]:
        """
        Updates the etags of watched settings that are part of the configuration
        :param List[ConfigurationSetting] configuration_settings: The list of configuration settings to update
        :return: A dictionary mapping (key, label) tuples to their updated etags
        :rtype: Dict[Tuple[str, str], Optional[str]]
        """
        watched_settings: Dict[Tuple[str, str], Optional[str]] = {}
        for config in configuration_settings:
            if (config.key, config.label) in self._watched_settings:
                watched_settings[(config.key, config.label)] = config.etag
        return watched_settings

    def _update_watched_feature_flags(
        self, feature_flags: List[FeatureFlagConfigurationSetting]
    ) -> Dict[Tuple[str, str], Optional[str]]:
        """
        Updates the etags of watched feature flags that are part of the configuration
        :param List[FeatureFlagConfigurationSetting] feature_flags: The list of feature flags to update
        :return: A dictionary mapping (key, label) tuples to their updated etags
        :rtype: Dict[Tuple[str, str], Optional[str]]
        """
        watched_feature_flags: Dict[Tuple[str, str], Optional[str]] = {}
        for feature_flag in feature_flags:
            watched_feature_flags[(feature_flag.key, feature_flag.label)] = feature_flag.etag
        return watched_feature_flags

    def _update_correlation_context_header(
        self,
        headers: Dict[str, str],
        request_type: str,
        replica_count: int,
        uses_key_vault: bool,
        is_failover_request: bool = False,
    ) -> Dict[str, str]:
        """
        Update the correlation context header with telemetry information.

        :param headers: The headers dictionary to update.
        :type headers: Dict[str, str]
        :param request_type: The type of request (e.g., "Startup", "Watch").
        :type request_type: str
        :param replica_count: The number of replica endpoints.
        :type replica_count: int
        :param uses_key_vault: Whether this request uses Key Vault.
        :type uses_key_vault: bool
        :param is_failover_request: Whether this is a failover request.
        :type is_failover_request: bool
        :return: The updated headers dictionary.
        :rtype: Dict[str, str]
        """
        return self._tracing_context.update_correlation_context_header(
            headers=headers,
            request_type=request_type,
            replica_count=replica_count,
            uses_key_vault=uses_key_vault,
            feature_flag_enabled=self._feature_flag_enabled,
            is_failover_request=is_failover_request,
        )

    def _deduplicate_settings(self, configuration_settings: List[ConfigurationSetting]) -> List[ConfigurationSetting]:
        """
        Deduplicates configuration settings by key.

        :param List[ConfigurationSetting] configuration_settings: The list of configuration settings to deduplicate
        :return: A list of unique configuration settings
        :rtype: List[ConfigurationSetting]
        """
        unique_settings: Dict[str, ConfigurationSetting] = {}
        for settings in configuration_settings:
            unique_settings[settings.key] = settings
        return list(unique_settings.values())



# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import unittest
import time
import datetime
import json
import base64
from unittest.mock import patch, Mock
from typing import Dict, Any

from azure.appconfiguration import FeatureFlagConfigurationSetting
from azure.appconfiguration.provider._azureappconfigurationproviderbase import (
    delay_failure,
    is_json_content_type,
    _build_watched_setting,
    sdk_allowed_kwargs,
    AzureAppConfigurationProviderBase,
)
from azure.appconfiguration.provider._models import SettingSelector
from azure.appconfiguration.provider._constants import (
    NULL_CHAR,
    TELEMETRY_KEY,
    METADATA_KEY,
    ETAG_KEY,
    FEATURE_FLAG_REFERENCE_KEY,
)
from azure.appconfiguration.provider._refresh_timer import _RefreshTimer


class TestDelayFailure(unittest.TestCase):
    """Test the delay_failure function."""

    def test_delay_failure_when_enough_time_passed(self):
        """Test that no delay occurs when enough time has passed."""
        start_time = datetime.datetime.now() - datetime.timedelta(seconds=10)
        with patch("time.sleep") as mock_sleep:
            delay_failure(start_time)
            mock_sleep.assert_not_called()

    def test_delay_failure_when_insufficient_time_passed(self):
        """Test that delay occurs when insufficient time has passed."""
        start_time = datetime.datetime.now() - datetime.timedelta(seconds=2)
        with patch("time.sleep") as mock_sleep:
            delay_failure(start_time)
            mock_sleep.assert_called_once()
            # Verify the delay is approximately correct (around 3 seconds)
            called_delay = mock_sleep.call_args[0][0]
            self.assertGreater(called_delay, 2)
            self.assertLess(called_delay, 4)


class TestIsJsonContentType(unittest.TestCase):
    """Test the is_json_content_type function."""

    def test_valid_json_content_types(self):
        """Test various valid JSON content types."""
        valid_types = [
            "application/json",
            "application/json; charset=utf-8",
            "APPLICATION/JSON",
            "application/vnd.api+json",
            "application/ld+json",
        ]
        for content_type in valid_types:
            with self.subTest(content_type=content_type):
                self.assertTrue(is_json_content_type(content_type))

    def test_invalid_json_content_types(self):
        """Test various invalid JSON content types."""
        invalid_types = [
            "",
            None,
            "text/plain",
            "application/xml",
            "text/json",  # Wrong main type
            "application",  # Malformed
            "application/",  # Malformed
        ]
        for content_type in invalid_types:
            with self.subTest(content_type=content_type):
                self.assertFalse(is_json_content_type(content_type))


class TestBuildWatchedSetting(unittest.TestCase):
    """Test the _build_watched_setting function."""

    def test_string_input(self):
        """Test with string input."""
        result = _build_watched_setting("test_key")
        self.assertEqual(result, ("test_key", NULL_CHAR))

    def test_tuple_input(self):
        """Test with tuple input."""
        result = _build_watched_setting(("test_key", "test_label"))
        self.assertEqual(result, ("test_key", "test_label"))

    def test_wildcard_key_raises_error(self):
        """Test that wildcard in key raises ValueError."""
        with self.assertRaises(ValueError):
            _build_watched_setting("test*key")

    def test_wildcard_label_raises_error(self):
        """Test that wildcard in label raises ValueError."""
        with self.assertRaises(ValueError):
            _build_watched_setting(("test_key", "test*label"))


class TestSdkAllowedKwargs(unittest.TestCase):
    """Test the sdk_allowed_kwargs function."""

    def test_filters_allowed_kwargs(self):
        """Test that only allowed kwargs are returned."""
        kwargs = {
            "headers": {"test": "value"},
            "timeout": 30,
            "invalid_param": "should_be_filtered",
            "user_agent": "test_agent",
            "unknown_param": "filtered_out",
        }
        result = sdk_allowed_kwargs(kwargs)
        expected = {"headers": {"test": "value"}, "timeout": 30, "user_agent": "test_agent"}
        self.assertEqual(result, expected)

    def test_empty_kwargs(self):
        """Test with empty kwargs."""
        result = sdk_allowed_kwargs({})
        self.assertEqual(result, {})


class TestRefreshTimer(unittest.TestCase):
    """Test the _RefreshTimer class."""

    def test_default_initialization(self):
        """Test default initialization."""
        timer = _RefreshTimer()
        self.assertEqual(timer._interval, 30)
        self.assertEqual(timer._attempts, 1)
        self.assertEqual(timer._min_backoff, 30)
        self.assertEqual(timer._max_backoff, 30)

    def test_custom_initialization(self):
        """Test custom initialization."""
        timer = _RefreshTimer(refresh_interval=60, min_backoff=10, max_backoff=300)
        self.assertEqual(timer._interval, 60)
        self.assertEqual(timer._min_backoff, 10)
        # max_backoff is constrained by the interval, so it should be 60, not 300
        self.assertEqual(timer._max_backoff, 60)

    def test_invalid_refresh_interval_raises_error(self):
        """Test that invalid refresh interval raises ValueError."""
        with self.assertRaises(ValueError):
            _RefreshTimer(refresh_interval=0)

    def test_needs_refresh_initially_false(self):
        """Test that needs_refresh is initially false."""
        timer = _RefreshTimer(refresh_interval=1)
        self.assertFalse(timer.needs_refresh())

    def test_needs_refresh_after_interval(self):
        """Test that needs_refresh becomes true after interval."""
        timer = _RefreshTimer(refresh_interval=1)
        # Manually set the next refresh time to the past
        timer._next_refresh_time = time.time() - 1
        self.assertTrue(timer.needs_refresh())

    def test_reset_functionality(self):
        """Test the reset functionality."""
        timer = _RefreshTimer(refresh_interval=1)
        timer._attempts = 5
        timer.reset()
        self.assertEqual(timer._attempts, 1)
        # Next refresh time should be reset to future
        self.assertGreater(timer._next_refresh_time, time.time())

    def test_backoff_increases_attempts(self):
        """Test that backoff increases attempts."""
        timer = _RefreshTimer(refresh_interval=60, min_backoff=1, max_backoff=60)
        initial_attempts = timer._attempts
        timer.backoff()
        self.assertEqual(timer._attempts, initial_attempts + 1)
        self.assertGreater(timer._next_refresh_time, time.time())

    def test_calculate_backoff_progression(self):
        """Test that backoff calculation progresses correctly."""
        timer = _RefreshTimer(refresh_interval=60, min_backoff=1, max_backoff=60)

        # Test multiple backoff calculations to verify exponential progression
        # Since backoff includes randomization, we test the range of possible values
        min_backoff_ms = 1000  # min_backoff in milliseconds
        max_backoff_ms = 60000  # max_backoff in milliseconds

        # For attempts=1, calculated value should be min_backoff * 2^1 = 2000ms
        # Random component can range from min_backoff (1000) to calculated (2000)
        backoff1 = timer._calculate_backoff()
        self.assertGreaterEqual(backoff1, min_backoff_ms)
        self.assertLessEqual(backoff1, 2000)  # min_backoff * 2^1

        timer._attempts += 1
        # For attempts=2, calculated value should be min_backoff * 2^2 = 4000ms
        # Random component can range from min_backoff (1000) to calculated (4000)
        backoff2 = timer._calculate_backoff()
        self.assertGreaterEqual(backoff2, min_backoff_ms)
        self.assertLessEqual(backoff2, 4000)  # min_backoff * 2^2

        # Both should be within overall min/max bounds
        self.assertLessEqual(backoff1, max_backoff_ms)
        self.assertLessEqual(backoff2, max_backoff_ms)


class TestAzureAppConfigurationProviderBase(unittest.TestCase):
    """Test the AzureAppConfigurationProviderBase class."""

    def setUp(self):
        """Set up test environment."""
        self.provider = AzureAppConfigurationProviderBase(endpoint="https://test.azconfig.io")

    def test_initialization_with_defaults(self):
        """Test initialization with default values."""
        provider = AzureAppConfigurationProviderBase()
        self.assertEqual(provider._origin_endpoint, "")
        self.assertEqual(provider._dict, {})
        self.assertIsInstance(provider._selects, list)
        self.assertEqual(len(provider._selects), 1)
        self.assertEqual(provider._trim_prefixes, [])
        self.assertFalse(provider._feature_flag_enabled)

    def test_initialization_with_custom_values(self):
        """Test initialization with custom values."""
        selects = [SettingSelector(key_filter="app:*")]
        trim_prefixes = ["app:", "config:"]
        refresh_on = ["refresh_key"]

        provider = AzureAppConfigurationProviderBase(
            endpoint="https://test.azconfig.io",
            selects=selects,
            trim_prefixes=trim_prefixes,
            refresh_on=refresh_on,
            feature_flag_enabled=True,
            refresh_interval=60,
        )

        self.assertEqual(provider._origin_endpoint, "https://test.azconfig.io")
        self.assertEqual(provider._selects, selects)
        self.assertEqual(provider._trim_prefixes, ["config:", "app:"])  # Should be sorted by length
        self.assertTrue(provider._feature_flag_enabled)
        self.assertEqual(provider._refresh_timer._interval, 60)

    def test_process_key_name_with_no_prefix(self):
        """Test key name processing with no matching prefix."""
        config = Mock()
        config.key = "test_key"

        result = self.provider._process_key_name(config)
        self.assertEqual(result, "test_key")

    def test_process_key_name_with_matching_prefix(self):
        """Test key name processing with matching prefix."""
        provider = AzureAppConfigurationProviderBase(trim_prefixes=["app:", "config:"])
        config = Mock()
        config.key = "app:test_key"

        result = provider._process_key_name(config)
        self.assertEqual(result, "test_key")

    def test_process_key_name_with_longest_matching_prefix(self):
        """Test key name processing uses longest matching prefix."""
        provider = AzureAppConfigurationProviderBase(trim_prefixes=["app:", "app:config:"])
        config = Mock()
        config.key = "app:config:test_key"

        result = provider._process_key_name(config)
        self.assertEqual(result, "test_key")

    def test_mapping_interface_empty(self):
        """Test mapping interface with empty provider."""
        self.assertEqual(len(self.provider), 0)
        self.assertNotIn("test_key", self.provider)
        self.assertEqual(list(self.provider.keys()), [])
        self.assertEqual(list(self.provider.values()), [])
        self.assertEqual(list(self.provider.items()), [])

    def test_mapping_interface_with_data(self):
        """Test mapping interface with data."""
        # Manually add data to test mapping interface
        with self.provider._update_lock:
            self.provider._dict = {"key1": "value1", "key2": {"nested": "value"}}

        self.assertEqual(len(self.provider), 2)
        self.assertIn("key1", self.provider)
        self.assertNotIn("key3", self.provider)
        self.assertEqual(self.provider["key1"], "value1")
        self.assertEqual(self.provider.get("key1"), "value1")
        self.assertEqual(self.provider.get("key3", "default"), "default")

        keys = list(self.provider.keys())
        self.assertIn("key1", keys)
        self.assertIn("key2", keys)

        values = list(self.provider.values())
        self.assertIn("value1", values)
        self.assertIn({"nested": "value"}, values)

    def test_process_key_value_base_plain_text(self):
        """Test processing non-keyvault plain text value."""
        config = Mock()
        config.content_type = "text/plain"
        config.value = "plain text value"

        result = self.provider._process_key_value_base(config)
        self.assertEqual(result, "plain text value")

    def test_process_key_value_base_json(self):
        """Test processing non-keyvault JSON value."""
        config = Mock()
        config.content_type = "application/json"
        config.value = '{"key": "value", "number": 42}'

        result = self.provider._process_key_value_base(config)
        expected = {"key": "value", "number": 42}
        self.assertEqual(result, expected)

    def test_process_key_value_base_invalid_json(self):
        """Test processing non-keyvault invalid JSON value."""
        config = Mock()
        config.content_type = "application/json"
        config.value = '{"invalid": json}'

        # Mock the remove_json_comments import to avoid dependency issues
        with patch("azure.appconfiguration.provider._azureappconfigurationproviderbase.json.loads") as mock_json_loads:
            mock_json_loads.side_effect = [
                json.JSONDecodeError("test", "test", 0),
                json.JSONDecodeError("test", "test", 0),
            ]
            result = self.provider._process_key_value_base(config)
            self.assertEqual(result, '{"invalid": json}')  # Should return as string

    def test_update_ff_telemetry_metadata(self):
        """Test feature flag telemetry processing."""
        feature_flag = Mock(spec=FeatureFlagConfigurationSetting)
        feature_flag.etag = "test_etag"
        feature_flag.key = "test_feature"
        feature_flag.label = "test_label"

        feature_flag_value: Dict[str, Any] = {TELEMETRY_KEY: {"enabled": True}}

        endpoint = "https://test.azconfig.io"

        self.provider._update_ff_telemetry_metadata(endpoint, feature_flag, feature_flag_value)

        # Verify telemetry structure was created
        self.assertIn(TELEMETRY_KEY, feature_flag_value)
        self.assertIn(METADATA_KEY, feature_flag_value[TELEMETRY_KEY])

        metadata = feature_flag_value[TELEMETRY_KEY][METADATA_KEY]
        self.assertEqual(metadata[ETAG_KEY], "test_etag")
        self.assertIn(FEATURE_FLAG_REFERENCE_KEY, metadata)
        self.assertIn("test_feature", metadata[FEATURE_FLAG_REFERENCE_KEY])

    def test_update_ff_telemetry_metadata_max_variants(self):
        """Test that max_variants only increases, never decreases."""
        feature_flag = FeatureFlagConfigurationSetting("test_feature")

        self.assertIsNone(self.provider._tracing_context.max_variants)

        feature_flag_value: Dict[str, Any] = {}

        self.provider._update_ff_telemetry_metadata("", feature_flag, feature_flag_value)

        # Verify max_variants remains None
        self.assertIsNone(self.provider._tracing_context.max_variants)

        # First call with 3 variants
        feature_flag_value_3: Dict[str, Any] = {"variants": [{}, {}, {}]}  # 3 variants

        self.provider._update_ff_telemetry_metadata("", feature_flag, feature_flag_value_3)
        self.assertEqual(self.provider._tracing_context.max_variants, 3)

        # Second call with 1 variant (should not decrease)
        feature_flag_value_1: Dict[str, Any] = {"variants": [{}]}  # 1 variant

        self.provider._update_ff_telemetry_metadata("", feature_flag, feature_flag_value_1)
        self.assertEqual(self.provider._tracing_context.max_variants, 3)  # Should remain 3

        # Third call with 5 variants (should increase)
        feature_flag_value_5: Dict[str, Any] = {"variants": [{}, {}, {}, {}, {}]}  # 5 variants

        self.provider._update_ff_telemetry_metadata("", feature_flag, feature_flag_value_5)
        self.assertEqual(self.provider._tracing_context.max_variants, 5)  # Should increase to 5

    def test_generate_allocation_id_no_allocation(self):
        """Test allocation ID generation with no allocation."""
        feature_flag_value: Dict[str, Any] = {"no_allocation": "here"}
        result = AzureAppConfigurationProviderBase._generate_allocation_id(feature_flag_value)
        self.assertIsNone(result)

    def test_generate_allocation_id_with_allocation(self):
        """Test allocation ID generation with allocation."""
        feature_flag_value: Dict[str, Any] = {
            "allocation": {
                "seed": "test_seed",
                "default_when_enabled": "Control",
                "percentile": [{"from": 0, "to": 50, "variant": "Control"}, {"from": 50, "to": 100, "variant": "Test"}],
            },
            "variants": [
                {"name": "Control", "configuration_value": {"key": "control_value"}},
                {"name": "Test", "configuration_value": {"key": "test_value"}},
            ],
        }

        result = AzureAppConfigurationProviderBase._generate_allocation_id(feature_flag_value)
        self.assertIsNotNone(result)
        self.assertIsInstance(result, str)
        # Should be a base64 encoded string
        if result is not None:
            try:
                base64.urlsafe_b64decode(result.encode() + b"==")  # Add padding if needed
            except Exception:
                self.fail("Result should be valid base64")

    def test_generate_allocation_id_no_variants_no_seed(self):
        """Test allocation ID generation with no variants and no seed."""
        feature_flag_value: Dict[str, Any] = {
            "allocation": {
                # Only having default_when_enabled means allocated_variants won't be empty
                # So this should actually return an allocation ID, not None
                "default_when_enabled": "Control"
            }
        }
        result = AzureAppConfigurationProviderBase._generate_allocation_id(feature_flag_value)
        # Since default_when_enabled is provided, allocated_variants won't be empty
        # so this should return a valid allocation ID
        self.assertIsNotNone(result)
        self.assertIsInstance(result, str)

    def test_generate_allocation_id_truly_empty(self):
        """Test allocation ID generation with truly empty allocation."""
        feature_flag_value: Dict[str, Any] = {
            "allocation": {
                # No seed and no default_when_enabled
            }
        }
        result = AzureAppConfigurationProviderBase._generate_allocation_id(feature_flag_value)
        # This should return None because allocated_variants is empty and no seed
        self.assertIsNone(result)



# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------

from ._azureappconfigurationprovider import load, AzureAppConfigurationProvider
from ._models import (
    AzureAppConfigurationKeyVaultOptions,
    SettingSelector,
    WatchKey,
)

from ._version import VERSION

__version__ = VERSION
__all__ = [
    "load",
    "AzureAppConfigurationProvider",
    "AzureAppConfigurationKeyVaultOptions",
    "SettingSelector",
    "WatchKey",
]



# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import os
import asyncio
from sample_utilities import get_client_modifications
from azure.appconfiguration.provider.aio import load
from azure.appconfiguration.provider import SettingSelector


async def main():
    kwargs = get_client_modifications()
    connection_string = os.environ["APPCONFIGURATION_CONNECTION_STRING"]

    # Connecting to Azure App Configuration using connection string
    config = await load(connection_string=connection_string, **kwargs)

    print(config["message"])
    print(config["my_json"]["key"])
    await config.close()

    # Connecting to Azure App Configuration using connection string and trimmed key prefixes
    trimmed = ["test."]
    config = await load(connection_string=connection_string, trim_prefixes=trimmed, **kwargs)

    print(config["message"])
    await config.close()

    # Connection to Azure App Configuration using SettingSelector
    selects = [SettingSelector(key_filter="message*")]
    config = await load(connection_string=connection_string, selects=selects, **kwargs)

    print("message found: " + str("message" in config))
    print("test.message found: " + str("test.message" in config))
    await config.close()


if __name__ == "__main__":
    asyncio.run(main())



# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import os
from typing import Dict, Optional, List, Tuple
from importlib.metadata import version, PackageNotFoundError
from ._constants import (
    REQUEST_TRACING_DISABLED_ENVIRONMENT_VARIABLE,
    ServiceFabricEnvironmentVariable,
    AzureFunctionEnvironmentVariable,
    AzureWebAppEnvironmentVariable,
    ContainerAppEnvironmentVariable,
    KubernetesEnvironmentVariable,
    APP_CONFIG_AI_MIME_PROFILE,
    APP_CONFIG_AICC_MIME_PROFILE,
)

# Feature flag filter names
PERCENTAGE_FILTER_NAMES = ["Percentage", "PercentageFilter", "Microsoft.Percentage", "Microsoft.PercentageFilter"]
TIME_WINDOW_FILTER_NAMES = ["TimeWindow", "TimeWindowFilter", "Microsoft.TimeWindow", "Microsoft.TimeWindowFilter"]
TARGETING_FILTER_NAMES = ["Targeting", "TargetingFilter", "Microsoft.Targeting", "Microsoft.TargetingFilter"]

CUSTOM_FILTER_KEY = "CSTM"  # cspell:disable-line
PERCENTAGE_FILTER_KEY = "PRCNT"  # cspell:disable-line
TIME_WINDOW_FILTER_KEY = "TIME"
TARGETING_FILTER_KEY = "TRGT"  # cspell:disable-line

FEATURE_FLAG_USES_TELEMETRY_TAG = "Telemetry"
FEATURE_FLAG_USES_SEED_TAG = "Seed"
FEATURE_FLAG_MAX_VARIANTS_KEY = "MaxVariants"
FEATURE_FLAG_FEATURES_KEY = "FFFeatures"

CHAT_COMPLETION_PROFILE = "chat-completion"


# Feature flag constants for telemetry
LOAD_BALANCING_FEATURE = "LB"
AI_CONFIGURATION_FEATURE = "AI"
AI_CHAT_COMPLETION_FEATURE = "AICC"
SNAPSHOT_REFERENCE_TAG = "SnapshotRef"

# Correlation context constants
FEATUREMANAGEMENT_PACKAGE = "featuremanagement"
CORRELATION_CONTEXT_HEADER = "Correlation-Context"
REQUEST_TYPE_KEY = "RequestType"
REPLICA_COUNT_KEY = "ReplicaCount"
HOST_KEY = "Host"
FILTER_KEY = "Filter"
MAX_VARIANTS_KEY = "MaxVariants"
FF_FEATURES_KEY = "FFFeatures"
FM_PY_VER_KEY = "FMPyVer"
FEATURES_KEY = "Features"
USES_KEY_VAULT_TAG = "UsesKeyVault"
FAILOVER_TAG = "Failover"


class HostType:
    UNIDENTIFIED = ""
    AZURE_WEB_APP = "AzureWebApp"
    AZURE_FUNCTION = "AzureFunction"
    CONTAINER_APP = "ContainerApp"
    KUBERNETES = "Kubernetes"
    SERVICE_FABRIC = "ServiceFabric"


class RequestType:
    STARTUP = "Startup"
    WATCH = "Watch"


Delimiter = "+"


class _RequestTracingContext:  # pylint: disable=too-many-instance-attributes
    """
    Encapsulates request tracing and telemetry configuration values.
    """

    def __init__(self, load_balancing_enabled: bool = False) -> None:
        # Main feature tracking properties
        self.uses_load_balancing = load_balancing_enabled
        self.uses_ai_configuration = False
        self.uses_aicc_configuration = False  # AI Chat Completion
        self.uses_snapshot_reference = False
        self.uses_telemetry = False
        self.uses_seed = False
        self.max_variants: Optional[int] = None
        self.feature_filter_usage: Dict[str, bool] = {}
        self.is_key_vault_configured: bool = False
        self.replica_count: int = 0
        self.is_failover_request: bool = False

        # Host and environment detection properties
        self.host_type: str = _RequestTracingContext.get_host_type()

        # Version tracking
        self.feature_management_version: Optional[str] = None

    def update_max_variants(self, size: int) -> None:
        """
        Update max_variants if the new size is larger.

        :param size: size of variant
        :type size: int
        """
        if self.max_variants is None or size > self.max_variants:
            self.max_variants = size

    def reset_ai_configuration_tracing(self) -> None:
        """
        Reset AI configuration tracing flags.
        """
        self.uses_ai_configuration = False
        self.uses_aicc_configuration = False

    def update_ai_configuration_tracing(self, content_type: Optional[str]) -> None:
        """
        Update AI configuration tracing based on content type.

        :param content_type: The content type to analyze.
        :type content_type: Optional[str]
        """
        if not content_type or self.uses_aicc_configuration:
            return

        # Check for AI mime profiles in content type
        if APP_CONFIG_AI_MIME_PROFILE in content_type:
            self.uses_ai_configuration = True
            if APP_CONFIG_AICC_MIME_PROFILE in content_type:
                self.uses_aicc_configuration = True

    def update_correlation_context_header(
        self,
        headers: Dict[str, str],
        request_type: str,
        replica_count: int,
        uses_key_vault: bool,
        feature_flag_enabled: bool,
        is_failover_request: bool = False,
    ) -> Dict[str, str]:
        """
        Update the correlation context header with telemetry information.

        :param headers: The headers dictionary to update.
        :type headers: Dict[str, str]
        :param request_type: The type of request (e.g., "Startup", "Watch").
        :type request_type: str
        :param replica_count: The number of replica endpoints.
        :type replica_count: int
        :param uses_key_vault: Whether this request uses Key Vault.
        :type uses_key_vault: bool
        :param feature_flag_enabled: Whether feature flags are enabled.
        :type feature_flag_enabled: bool
        :param is_failover_request: Whether this is a failover request.
        :type is_failover_request: bool
        :return: The updated headers dictionary.
        :rtype: Dict[str, str]
        """
        if os.environ.get(REQUEST_TRACING_DISABLED_ENVIRONMENT_VARIABLE, "").lower() == "true":
            return headers

        # Update instance properties for the correlation context
        self.replica_count = replica_count
        self.is_key_vault_configured = uses_key_vault
        self.is_failover_request = is_failover_request

        # Update feature management version if needed and feature flags are enabled
        if feature_flag_enabled and not self.feature_management_version:
            self.feature_management_version = self.get_assembly_version(FEATUREMANAGEMENT_PACKAGE)

        # Key-value pairs for the correlation context
        key_values: List[Tuple[str, str]] = []
        tags: List[str] = []

        # Add request type
        key_values.append((REQUEST_TYPE_KEY, request_type))

        # Add replica count if configured
        if self.replica_count > 0:
            key_values.append((REPLICA_COUNT_KEY, str(self.replica_count)))

        # Add host type if identified
        if self.host_type != HostType.UNIDENTIFIED:
            key_values.append((HOST_KEY, self.host_type))

        # Add feature filter information
        if len(self.feature_filter_usage) > 0:
            filters_string = self._create_filters_string()
            if filters_string:
                key_values.append((FILTER_KEY, filters_string))

        # Add max variants if present
        if self.max_variants and self.max_variants > 0:
            key_values.append((MAX_VARIANTS_KEY, str(self.max_variants)))

        # Add feature flag features if present
        ff_features_string = self._create_ff_features_string()
        if ff_features_string:
            key_values.append((FF_FEATURES_KEY, ff_features_string))

        # Add general features if present
        features_string = self._create_features_string()
        if features_string:
            key_values.append((FEATURES_KEY, features_string))

        # Add version information
        if self.feature_management_version:
            key_values.append((FM_PY_VER_KEY, self.feature_management_version))

        # Add tags
        if self.is_key_vault_configured:
            tags.append(USES_KEY_VAULT_TAG)

        if self.is_failover_request:
            tags.append(FAILOVER_TAG)

        # Build the correlation context string
        context_parts: List[str] = []

        # Add key-value pairs
        for key, value in key_values:
            context_parts.append(f"{key}={value}")

        # Add tags
        context_parts.extend(tags)

        correlation_context = ",".join(context_parts)

        if correlation_context:
            headers[CORRELATION_CONTEXT_HEADER] = correlation_context

        return headers

    def update_feature_filter_telemetry(self, feature_flag) -> None:
        """
        Track feature filter usage for App Configuration telemetry.

        :param feature_flag: The feature flag to analyze for filter usage.
        :type feature_flag: FeatureFlagConfigurationSetting
        """
        # Constants are already imported at module level

        if feature_flag.filters:
            for filter in feature_flag.filters:
                if filter.get("name") in PERCENTAGE_FILTER_NAMES:
                    self.feature_filter_usage[PERCENTAGE_FILTER_KEY] = True
                elif filter.get("name") in TIME_WINDOW_FILTER_NAMES:
                    self.feature_filter_usage[TIME_WINDOW_FILTER_KEY] = True
                elif filter.get("name") in TARGETING_FILTER_NAMES:
                    self.feature_filter_usage[TARGETING_FILTER_KEY] = True
                else:
                    self.feature_filter_usage[CUSTOM_FILTER_KEY] = True

    def reset_feature_filter_usage(self) -> None:
        """Reset the feature filter usage tracking."""
        self.feature_filter_usage = {}

    def _create_features_string(self) -> str:
        """
        Generate the features string for correlation context.

        :return: A string containing features used separated by delimiters.
        :rtype: str
        """
        features_list = []

        if self.uses_load_balancing:
            features_list.append(LOAD_BALANCING_FEATURE)
        if self.uses_ai_configuration:
            features_list.append(AI_CONFIGURATION_FEATURE)
        if self.uses_aicc_configuration:
            features_list.append(AI_CHAT_COMPLETION_FEATURE)
        if self.uses_snapshot_reference:
            features_list.append(SNAPSHOT_REFERENCE_TAG)

        return Delimiter.join(features_list)

    def _create_ff_features_string(self) -> str:
        """
        Generate the features string for feature flag usage tracking.

        :return: A string containing feature flag usage tags separated by delimiters.
        :rtype: str
        """
        features_list = []

        if self.uses_seed:
            features_list.append(FEATURE_FLAG_USES_SEED_TAG)

        if self.uses_telemetry:
            features_list.append(FEATURE_FLAG_USES_TELEMETRY_TAG)

        return Delimiter.join(features_list)

    def _create_filters_string(self) -> str:
        """
        Create a string representing the feature filters in use.

        :return: String of filter names separated by delimiters.
        :rtype: str
        """
        filters: List[str] = []

        if CUSTOM_FILTER_KEY in self.feature_filter_usage:
            filters.append(CUSTOM_FILTER_KEY)
        if PERCENTAGE_FILTER_KEY in self.feature_filter_usage:
            filters.append(PERCENTAGE_FILTER_KEY)
        if TIME_WINDOW_FILTER_KEY in self.feature_filter_usage:
            filters.append(TIME_WINDOW_FILTER_KEY)
        if TARGETING_FILTER_KEY in self.feature_filter_usage:
            filters.append(TARGETING_FILTER_KEY)

        return Delimiter.join(filters)

    @staticmethod
    def get_host_type() -> str:
        """
        Detect the host environment type based on environment variables.

        :return: The detected host type.
        :rtype: str
        """
        if os.environ.get(AzureFunctionEnvironmentVariable):
            return HostType.AZURE_FUNCTION
        if os.environ.get(AzureWebAppEnvironmentVariable):
            return HostType.AZURE_WEB_APP
        if os.environ.get(ContainerAppEnvironmentVariable):
            return HostType.CONTAINER_APP
        if os.environ.get(KubernetesEnvironmentVariable):
            return HostType.KUBERNETES
        if os.environ.get(ServiceFabricEnvironmentVariable):
            return HostType.SERVICE_FABRIC

        return HostType.UNIDENTIFIED

    @staticmethod
    def get_assembly_version(package_name: str) -> Optional[str]:
        """
        Get the version of a Python package.

        :param package_name: The name of the package to get version for.
        :type package_name: str
        :return: Package version string or None if not found.
        :rtype: Optional[str]
        """
        if not package_name:
            return None

        try:
            return version(package_name)
        except PackageNotFoundError:
            pass

        return None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import time
import asyncio
import unittest
from unittest.mock import Mock, patch
from devtools_testutils.aio import recorded_by_proxy_async
from async_preparers import app_config_aad_decorator_async
from asynctestcase import AppConfigTestCase
from azure.appconfiguration import SecretReferenceConfigurationSetting
from azure.appconfiguration.provider import SettingSelector, WatchKey


class TestAsyncSecretRefresh(AppConfigTestCase, unittest.TestCase):

    @app_config_aad_decorator_async
    @recorded_by_proxy_async
    async def testsecret_refresh_timer(
        self,
        appconfiguration_endpoint_string,
        appconfiguration_keyvault_secret_url,
        appconfiguration_keyvault_secret_url2,
    ):
        """Test that secrets are refreshed based on the secret_refresh_interval."""

        # Create an async mock callback
        async def async_callback():
            pass

        mock_callback = Mock(side_effect=async_callback)

        # Create client with key vault reference and secret refresh interval
        client = await self.create_client(
            endpoint=appconfiguration_endpoint_string,
            selects={SettingSelector(key_filter="*", label_filter="prod")},
            keyvault_secret_url=appconfiguration_keyvault_secret_url,
            keyvault_secret_url2=appconfiguration_keyvault_secret_url2,
            on_refresh_success=mock_callback,
            refresh_interval=999999,
            secret_refresh_interval=1,
        )

        # Verify initial state
        assert client["secret"] == "Very secret value"
        assert mock_callback.call_count == 0

        # Mock the refresh method to track calls
        with patch.object(client, "refresh") as mock_refresh:
            # Wait for the secret refresh interval to pass
            await asyncio.sleep(2)

            await client.refresh()

            # Verify refresh was called
            assert mock_refresh.call_count >= 1

            # Wait again to ensure multiple refreshes
            await asyncio.sleep(2)
            await client.refresh()

            # Should have been called at least twice now
            assert mock_refresh.call_count >= 2

    @app_config_aad_decorator_async
    @recorded_by_proxy_async
    async def test_secret_refresh_with_updated_values(
        self,
        appconfiguration_endpoint_string,
        appconfiguration_keyvault_secret_url,
        appconfiguration_keyvault_secret_url2,
    ):
        """Test that secrets are refreshed with updated values."""
        mock_callback = Mock()

        # Create client with the mock secret resolver
        client = await self.create_client(
            endpoint=appconfiguration_endpoint_string,
            selects={SettingSelector(key_filter="*", label_filter="prod")},
            keyvault_secret_url=appconfiguration_keyvault_secret_url,
            keyvault_secret_url2=appconfiguration_keyvault_secret_url2,
            on_refresh_success=mock_callback,
            refresh_on=[WatchKey("secret", "prod")],
            refresh_interval=1,
            secret_refresh_interval=1,  # Using a short interval for testing
        )

        # Add a key vault reference to the client (this will use mock resolver)
        appconfig_client = self.create_aad_sdk_client(appconfiguration_endpoint_string)

        # Get and modify a key vault reference setting
        kv_setting = await appconfig_client.get_configuration_setting(key="secret", label="prod")
        assert kv_setting is not None

        # Verify initial value from mock resolver
        assert client["secret"] == "Very secret value"
        assert kv_setting is not None
        assert isinstance(kv_setting, SecretReferenceConfigurationSetting)
        # Update the secret_id (which is the value for SecretReferenceConfigurationSetting)
        kv_setting.secret_id = appconfiguration_keyvault_secret_url2
        await appconfig_client.set_configuration_setting(kv_setting)

        # Wait for the secret refresh interval to pass
        await asyncio.sleep(2)

        # Access the value again to trigger refresh
        await client.refresh()

        # Verify the value was updated
        assert client["secret"] == "Very secret value 2"
        assert mock_callback.call_count >= 1

    @app_config_aad_decorator_async
    @recorded_by_proxy_async
    async def test_no_secret_refresh_without_timer(
        self,
        appconfiguration_endpoint_string,
        appconfiguration_keyvault_secret_url,
        appconfiguration_keyvault_secret_url2,
    ):
        """Test that secrets are not refreshed if secret_refresh_interval is not set."""

        # Create an async mock callback
        async def async_callback():
            pass

        mock_callback = Mock(side_effect=async_callback)

        # Create client without specifying secret_refresh_interval
        client = await self.create_client(
            endpoint=appconfiguration_endpoint_string,
            selects={SettingSelector(key_filter="*", label_filter="prod")},
            keyvault_secret_url=appconfiguration_keyvault_secret_url,
            keyvault_secret_url2=appconfiguration_keyvault_secret_url2,
            on_refresh_success=mock_callback,
            refresh_interval=999999,
        )

        # Verify initial state
        assert client["secret"] == "Very secret value"

        # Mock the refresh method to track calls
        with patch("time.time") as mock_time:
            # Make time.time() return increasing values to simulate passage of time
            mock_time.side_effect = [time.time(), time.time() + 100]

            # Access the key vault reference - this shouldn't trigger an auto-refresh since
            # we didn't set a secret_refresh_interval
            await client.refresh()

            # Access it again to verify no auto-refresh due to secrets timer
            await client.refresh()

            # The mock_time should have been called twice (for our side_effect setup)
            # but there should be no automatic refresh caused by the secret timer
            assert mock_time.call_count == 2

    @app_config_aad_decorator_async
    @recorded_by_proxy_async
    async def test_secret_refresh_timer_triggers_refresh(
        self,
        appconfiguration_endpoint_string,
        appconfiguration_keyvault_secret_url,
        appconfiguration_keyvault_secret_url2,
    ):
        """Test that the secret refresh timer triggers a refresh after the specified interval."""

        # Create an async mock callback
        async def async_callback():
            pass

        mock_callback = Mock(side_effect=async_callback)

        # Create client with key vault reference and separate refresh intervals
        client = await self.create_client(
            endpoint=appconfiguration_endpoint_string,
            selects={SettingSelector(key_filter="*", label_filter="prod")},
            keyvault_secret_url=appconfiguration_keyvault_secret_url,
            keyvault_secret_url2=appconfiguration_keyvault_secret_url2,
            on_refresh_success=mock_callback,
            refresh_interval=999999,
            secret_refresh_interval=5,  # Secret refresh interval is short
        )

        # Now patch the refresh method and secret_refresh_timer to control behavior
        with patch.object(client, "refresh") as mock_refresh:
            # Now patch the secret_refresh_timer to control its behavior
            with patch.object(client._secret_provider, "secret_refresh_timer") as mock_timer:
                # Make needs_refresh() return True to simulate timer expiration
                mock_timer.needs_refresh.return_value = True

                # Access a key vault reference which should trigger refresh due to timer
                await client.refresh()

                # Verify refresh was called
                assert mock_refresh.call_count > 0

    @app_config_aad_decorator_async
    @recorded_by_proxy_async
    async def test_secret_refresh_interval_parameter(
        self,
        appconfiguration_endpoint_string,
        appconfiguration_keyvault_secret_url,
        appconfiguration_keyvault_secret_url2,
    ):
        """Test that secret_refresh_interval parameter is correctly passed and used."""

        # Create an async mock callback
        async def async_callback():
            pass

        mock_callback = Mock(side_effect=async_callback)

        # Create client with specific secret_refresh_interval
        client = await self.create_client(
            endpoint=appconfiguration_endpoint_string,
            selects={SettingSelector(key_filter="*", label_filter="prod")},
            keyvault_secret_url=appconfiguration_keyvault_secret_url,
            keyvault_secret_url2=appconfiguration_keyvault_secret_url2,
            on_refresh_success=mock_callback,
            refresh_interval=999999,
            secret_refresh_interval=42,  # Use a specific value we can check for
        )

        # Verify the secret refresh timer exists
        assert client._secret_provider.secret_refresh_timer is not None

        # We can only verify that it exists, but can't directly access the internal refresh_interval
        # as it's a protected attribute

        # Check with no refresh interval to ensure it's properly handled
        client2 = await self.create_client(
            endpoint=appconfiguration_endpoint_string,
            selects={SettingSelector(key_filter="*", label_filter="prod")},
            keyvault_secret_url=appconfiguration_keyvault_secret_url,
            keyvault_secret_url2=appconfiguration_keyvault_secret_url2,
            on_refresh_success=mock_callback,
            # No secret_refresh_interval specified
        )

        # Verify timer is created only when secret_refresh_interval is provided
        assert client._secret_provider.secret_refresh_timer is not None
        assert client2._secret_provider.secret_refresh_timer is None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import unittest
from unittest.mock import Mock, patch, AsyncMock
from devtools_testutils.aio import recorded_by_proxy_async
from async_preparers import app_config_decorator_async
from asynctestcase import AppConfigTestCase
from azure.appconfiguration import SecretReferenceConfigurationSetting
from azure.keyvault.secrets.aio import SecretClient
from azure.appconfiguration.provider.aio._key_vault._async_secret_provider import SecretProvider

TEST_SECRET_ID = "https://myvault.vault.azure.net/secrets/my_secret"

TEST_SECRET_ID_VERSION = TEST_SECRET_ID + "/12345"


class TestSecretProviderAsync(AppConfigTestCase, unittest.IsolatedAsyncioTestCase):

    async def test_init_with_defaults(self):
        """Test initialization of SecretProvider with default parameters."""
        secret_provider = SecretProvider()

        # Verify initialization with defaults
        self.assertEqual(len(secret_provider._secret_clients), 0)
        self.assertIsNone(secret_provider._keyvault_credential)
        self.assertIsNone(secret_provider._secret_resolver)
        self.assertEqual(secret_provider._keyvault_client_configs, {})
        self.assertFalse(secret_provider.uses_key_vault)
        self.assertIsNone(secret_provider.secret_refresh_timer)
        self.assertEqual(len(secret_provider._secret_cache), 0)

    async def test_init_with_keyvault_credential(self):
        """Test initialization with a Key Vault credential."""
        mock_credential = Mock()
        secret_provider = SecretProvider(keyvault_credential=mock_credential)

        # Verify initialization with a Key Vault credential
        self.assertEqual(len(secret_provider._secret_clients), 0)
        self.assertEqual(secret_provider._keyvault_credential, mock_credential)
        self.assertIsNone(secret_provider._secret_resolver)
        self.assertEqual(secret_provider._keyvault_client_configs, {})
        self.assertTrue(secret_provider.uses_key_vault)

    async def test_init_with_secret_resolver(self):
        """Test initialization with a secret resolver."""
        mock_resolver = Mock()
        secret_provider = SecretProvider(secret_resolver=mock_resolver)

        # Verify initialization with a secret resolver
        self.assertEqual(len(secret_provider._secret_clients), 0)
        self.assertIsNone(secret_provider._keyvault_credential)
        self.assertEqual(secret_provider._secret_resolver, mock_resolver)
        self.assertEqual(secret_provider._keyvault_client_configs, {})
        self.assertTrue(secret_provider.uses_key_vault)

    async def test_init_with_keyvault_client_configs(self):
        """Test initialization with Key Vault client configurations."""
        client_configs = {"https://myvault.vault.azure.net/": {"retry_total": 3}}
        secret_provider = SecretProvider(keyvault_client_configs=client_configs)

        # Verify initialization with Key Vault client configurations
        self.assertEqual(len(secret_provider._secret_clients), 0)
        self.assertIsNone(secret_provider._keyvault_credential)
        self.assertIsNone(secret_provider._secret_resolver)
        self.assertEqual(secret_provider._keyvault_client_configs, client_configs)
        self.assertTrue(secret_provider.uses_key_vault)

    async def test_init_with_secret_refresh_interval(self):
        """Test initialization with a secret refresh interval."""
        mock_credential = Mock()
        refresh_interval = 30
        secret_provider = SecretProvider(keyvault_credential=mock_credential, secret_refresh_interval=refresh_interval)

        # Verify initialization with a secret refresh interval
        self.assertIsNotNone(secret_provider.secret_refresh_timer)
        self.assertTrue(secret_provider.uses_key_vault)

    async def test_resolve_keyvault_reference_with_cached_secret(self):
        """Test resolving a Key Vault reference when the secret is in the cache."""
        # Create a mock Key Vault reference
        config = SecretReferenceConfigurationSetting(key="test-key", secret_id=TEST_SECRET_ID)

        # Create a SecretProvider with a mock credential
        secret_provider = SecretProvider(keyvault_credential=Mock())
        key_vault_identifier, _ = secret_provider.resolve_keyvault_reference_base(config)

        # Add to cache
        secret_provider._secret_cache[key_vault_identifier.source_id] = (
            key_vault_identifier,
            "test-key",
            "cached-secret-value",
        )

        # This should return the cached value without calling SecretClient
        result = await secret_provider.resolve_keyvault_reference(config)

        # Verify the result
        self.assertEqual(result, "cached-secret-value")

    async def test_resolve_keyvault_reference_with_cached_secret_version(self):
        """Test resolving a Key Vault reference when the secret is in the cache."""
        # Create a mock Key Vault reference
        config = SecretReferenceConfigurationSetting(key="test-key", secret_id=TEST_SECRET_ID_VERSION)

        # Create a SecretProvider with a mock credential
        secret_provider = SecretProvider(keyvault_credential=Mock())
        key_vault_identifier, _ = secret_provider.resolve_keyvault_reference_base(config)

        # Add to cache
        secret_provider._secret_cache[key_vault_identifier.source_id] = (
            key_vault_identifier,
            "test-key",
            "cached-secret-value",
        )

        # This should return the cached value without calling SecretClient
        result = await secret_provider.resolve_keyvault_reference(config)

        # Verify the result
        self.assertEqual(result, "cached-secret-value")

    async def test_resolve_keyvault_reference_with_existing_client(self):
        """Test resolving a Key Vault reference with an existing client."""
        # Create a mock Key Vault reference
        config = SecretReferenceConfigurationSetting(key="test-key", secret_id=TEST_SECRET_ID_VERSION)

        # Create a SecretProvider with a mock credential
        mock_credential = Mock()
        secret_provider = SecretProvider(keyvault_credential=mock_credential)

        # Create a mock SecretClient
        mock_client = Mock()
        mock_secret = Mock()
        mock_secret.value = "secret-value"
        # Set up the get_secret method to return an awaitable that resolves to mock_secret
        mock_client.get_secret = AsyncMock(return_value=mock_secret)

        # Add the mock client to the secret_clients dictionary
        vault_url = "https://myvault.vault.azure.net/"
        secret_provider._secret_clients[vault_url] = mock_client

        # Setup key vault identifier mock
        with patch("azure.keyvault.secrets.KeyVaultSecretIdentifier") as mock_kv_id:
            mock_id_instance = Mock()
            mock_id_instance._resource_id = TEST_SECRET_ID_VERSION
            mock_id_instance.source_id = TEST_SECRET_ID_VERSION
            mock_id_instance.name = "my_secret"
            mock_id_instance.version = "12345"
            mock_id_instance.vault_url = "https://myvault.vault.azure.net"
            mock_kv_id.return_value = mock_id_instance

            # Call resolve_keyvault_reference
            with patch.object(secret_provider, "resolve_keyvault_reference_base") as mock_base:
                mock_base.return_value = (mock_id_instance, vault_url)

                result = await secret_provider.resolve_keyvault_reference(config)

                # Verify the result
                self.assertEqual(result, "secret-value")
                mock_client.get_secret.assert_called_once_with(mock_id_instance.name, version=mock_id_instance.version)
                # Verify the secret was cached
                _, _, value = secret_provider._secret_cache[TEST_SECRET_ID_VERSION]
                self.assertEqual(value, "secret-value")

    async def test_resolve_keyvault_reference_with_new_client(self):
        """Test resolving a Key Vault reference by creating a new client."""
        # Create a mock Key Vault reference
        config = SecretReferenceConfigurationSetting(key="test-key", secret_id=TEST_SECRET_ID_VERSION)

        # Create a SecretProvider with a mock credential
        mock_credential = Mock()
        secret_provider = SecretProvider(keyvault_credential=mock_credential)

        # Setup key vault identifier mock
        with patch("azure.keyvault.secrets.KeyVaultSecretIdentifier") as mock_kv_id:
            mock_id_instance = Mock()
            mock_id_instance._resource_id = TEST_SECRET_ID_VERSION
            mock_id_instance.source_id = TEST_SECRET_ID_VERSION
            mock_id_instance.name = "my_secret"
            mock_id_instance.version = "12345"
            mock_id_instance.vault_url = "https://myvault.vault.azure.net"
            mock_kv_id.return_value = mock_id_instance

            # Mock SecretClient creation and get_secret method
            with patch(
                "azure.appconfiguration.provider.aio._key_vault._async_secret_provider.SecretClient"
            ) as mock_client_class:
                mock_client = Mock()
                mock_secret = Mock()
                mock_secret.value = "new-secret-value"
                mock_client.get_secret = AsyncMock(return_value=mock_secret)
                mock_client_class.return_value = mock_client

                # Call resolve_keyvault_reference
                with patch.object(secret_provider, "resolve_keyvault_reference_base") as mock_base:
                    vault_url = "https://myvault.vault.azure.net/"
                    mock_base.return_value = (mock_id_instance, vault_url)

                    result = await secret_provider.resolve_keyvault_reference(config)

                    # Verify the result
                    self.assertEqual(result, "new-secret-value")
                    mock_client_class.assert_called_once_with(vault_url=vault_url, credential=mock_credential)
                    mock_client.get_secret.assert_called_once_with(
                        mock_id_instance.name, version=mock_id_instance.version
                    )
                    # Verify the client was cached
                    self.assertEqual(secret_provider._secret_clients[vault_url], mock_client)
                    # Verify the secret was cached
                    _, _, value = secret_provider._secret_cache[TEST_SECRET_ID_VERSION]
                    self.assertEqual(value, "new-secret-value")

    async def test_resolve_keyvault_reference_with_secret_resolver(self):
        """Test resolving a Key Vault reference using a secret resolver."""
        # Create a mock Key Vault reference
        config = SecretReferenceConfigurationSetting(key="test-key", secret_id=TEST_SECRET_ID_VERSION)

        # Create a mock secret resolver
        mock_resolver = Mock(return_value="resolved-secret-value")

        # Create a SecretProvider with the mock resolver
        secret_provider = SecretProvider(secret_resolver=mock_resolver)

        # Setup key vault identifier mock
        with patch("azure.keyvault.secrets.KeyVaultSecretIdentifier") as mock_kv_id:
            mock_id_instance = Mock()
            mock_id_instance._resource_id = TEST_SECRET_ID_VERSION
            mock_id_instance.source_id = TEST_SECRET_ID_VERSION
            mock_id_instance.name = "my_secret"
            mock_id_instance.version = "12345"
            mock_id_instance.vault_url = "https://myvault.vault.azure.net"
            mock_kv_id.return_value = mock_id_instance

            # Call resolve_keyvault_reference
            with patch.object(secret_provider, "resolve_keyvault_reference_base") as mock_base:
                vault_url = "https://myvault.vault.azure.net/"
                mock_base.return_value = (mock_id_instance, vault_url)

                result = await secret_provider.resolve_keyvault_reference(config)

                # Verify the result
                self.assertEqual(result, "resolved-secret-value")
                mock_resolver.assert_called_once_with(TEST_SECRET_ID_VERSION)
                # Verify the secret was cached
                _, _, value = secret_provider._secret_cache[TEST_SECRET_ID_VERSION]
                self.assertEqual(value, "resolved-secret-value")

    async def test_resolve_keyvault_reference_with_async_secret_resolver(self):
        """Test resolving a Key Vault reference using an async secret resolver."""
        # Create a mock Key Vault reference
        config = SecretReferenceConfigurationSetting(key="test-key", secret_id=TEST_SECRET_ID_VERSION)

        # Create a mock async secret resolver
        async def async_resolver(secret_id):
            return "async-resolved-secret-value"

        # Create a SecretProvider with the mock resolver
        secret_provider = SecretProvider(secret_resolver=async_resolver)

        # Setup key vault identifier mock
        with patch("azure.keyvault.secrets.KeyVaultSecretIdentifier") as mock_kv_id:
            mock_id_instance = Mock()
            mock_id_instance._resource_id = TEST_SECRET_ID_VERSION
            mock_id_instance.source_id = TEST_SECRET_ID_VERSION
            mock_id_instance.name = "my_secret"
            mock_id_instance.version = "12345"
            mock_id_instance.vault_url = "https://myvault.vault.azure.net"
            mock_kv_id.return_value = mock_id_instance

            # Call resolve_keyvault_reference
            with patch.object(secret_provider, "resolve_keyvault_reference_base") as mock_base:
                vault_url = "https://myvault.vault.azure.net/"
                mock_base.return_value = (mock_id_instance, vault_url)

                result = await secret_provider.resolve_keyvault_reference(config)

                # Verify the result
                self.assertEqual(result, "async-resolved-secret-value")
                # Verify the secret was cached
                _, _, value = secret_provider._secret_cache[TEST_SECRET_ID_VERSION]
                self.assertEqual(value, "async-resolved-secret-value")

    async def test_resolve_keyvault_reference_with_client_and_resolver_fallback(self):
        """Test falling back to a secret resolver if the client fails to get the secret."""
        # Create a mock Key Vault reference
        config = SecretReferenceConfigurationSetting(key="test-key", secret_id=TEST_SECRET_ID_VERSION)

        # Create a mock credential and secret resolver
        mock_credential = Mock()
        mock_resolver = Mock(return_value="fallback-secret-value")

        # Create a SecretProvider with both credential and resolver
        secret_provider = SecretProvider(keyvault_credential=mock_credential, secret_resolver=mock_resolver)

        # Create a mock SecretClient that returns None for get_secret
        mock_client = Mock()
        mock_secret = Mock()
        mock_secret.value = None
        mock_client.get_secret = AsyncMock(return_value=mock_secret)

        # Add the mock client to the secret_clients dictionary
        vault_url = "https://myvault.vault.azure.net/"
        secret_provider._secret_clients[vault_url] = mock_client

        # Setup key vault identifier mock
        with patch("azure.keyvault.secrets.KeyVaultSecretIdentifier") as mock_kv_id:
            mock_id_instance = Mock()
            mock_id_instance._resource_id = TEST_SECRET_ID_VERSION
            mock_id_instance.source_id = TEST_SECRET_ID_VERSION
            mock_id_instance.name = "my_secret"
            mock_id_instance.version = "12345"
            mock_id_instance.vault_url = "https://myvault.vault.azure.net"
            mock_kv_id.return_value = mock_id_instance

            # Call resolve_keyvault_reference
            with patch.object(secret_provider, "resolve_keyvault_reference_base") as mock_base:
                mock_base.return_value = (mock_id_instance, vault_url)

                result = await secret_provider.resolve_keyvault_reference(config)

                # Verify the result
                self.assertEqual(result, "fallback-secret-value")
                mock_client.get_secret.assert_called_once_with(mock_id_instance.name, version=mock_id_instance.version)
                mock_resolver.assert_called_once_with(TEST_SECRET_ID_VERSION)
                # Verify the secret was cached
                _, _, value = secret_provider._secret_cache[TEST_SECRET_ID_VERSION]
                self.assertEqual(value, "fallback-secret-value")

    async def test_resolve_keyvault_reference_no_client_no_resolver(self):
        """Test that an error is raised when no client or resolver can resolve the reference."""
        # Create a mock Key Vault reference
        config = SecretReferenceConfigurationSetting(key="test-key", secret_id=TEST_SECRET_ID_VERSION)

        # Create a SecretProvider with a credential but no clients or resolvers
        mock_credential = Mock()
        secret_provider = SecretProvider(keyvault_credential=mock_credential)

        # Setup key vault identifier mock
        with patch("azure.keyvault.secrets.KeyVaultSecretIdentifier") as mock_kv_id:
            mock_id_instance = Mock()
            mock_id_instance._resource_id = TEST_SECRET_ID_VERSION
            mock_id_instance.source_id = TEST_SECRET_ID_VERSION
            mock_id_instance.name = "my_secret"
            mock_id_instance.version = "12345"
            mock_id_instance.vault_url = "https://myvault.vault.azure.net"
            mock_kv_id.return_value = mock_id_instance

            # Call resolve_keyvault_reference
            with patch.object(secret_provider, "resolve_keyvault_reference_base") as mock_base:
                mock_base.return_value = (mock_id_instance, "https://othervault.vault.azure.net/")

                # This should raise an error since we have no client for this vault URL
                with self.assertRaises(ValueError):
                    await secret_provider.resolve_keyvault_reference(config)
                await secret_provider.close()

    async def test_close(self):
        """Test closing the SecretProvider."""
        # Create a SecretProvider with mock clients
        secret_provider = SecretProvider()

        # Create mock clients
        mock_client1 = Mock()
        mock_client1.close = AsyncMock()
        mock_client2 = Mock()
        mock_client2.close = AsyncMock()

        # Add the mock clients to the secret_clients dictionary
        secret_provider._secret_clients = {
            "https://vault1.vault.azure.net/": mock_client1,
            "https://vault2.vault.azure.net/": mock_client2,
        }

        # Call close
        await secret_provider.close()

        # Verify both clients were closed
        mock_client1.close.assert_called_once()
        mock_client2.close.assert_called_once()

    async def test_client_config_specific_credential(self):
        """Test that client configuration can specify a specific credential."""
        # Create a mock Key Vault reference
        config = SecretReferenceConfigurationSetting(key="test-key", secret_id=TEST_SECRET_ID_VERSION)

        # Create mock credentials
        mock_default_credential = Mock(name="default_credential")
        mock_specific_credential = Mock(name="specific_credential")

        # Create client configs with a specific credential
        client_configs = {
            "https://myvault.vault.azure.net/": {"credential": mock_specific_credential, "retry_total": 3}
        }

        # Create a SecretProvider with default credential and client configs
        secret_provider = SecretProvider(
            keyvault_credential=mock_default_credential, keyvault_client_configs=client_configs
        )

        # Setup key vault identifier mock
        with patch("azure.keyvault.secrets.KeyVaultSecretIdentifier") as mock_kv_id:
            mock_id_instance = Mock()
            mock_id_instance._resource_id = TEST_SECRET_ID_VERSION
            mock_id_instance.source_id = TEST_SECRET_ID_VERSION
            mock_id_instance.name = "my_secret"
            mock_id_instance.version = "12345"
            mock_id_instance.vault_url = "https://myvault.vault.azure.net"
            mock_kv_id.return_value = mock_id_instance

            # Mock SecretClient creation and get_secret method
            with patch(
                "azure.appconfiguration.provider.aio._key_vault._async_secret_provider.SecretClient"
            ) as mock_client_class:
                mock_client = Mock()
                mock_secret = Mock()
                mock_secret.value = "secret-value"
                mock_client.get_secret = AsyncMock(return_value=mock_secret)
                mock_client_class.return_value = mock_client

                # Call resolve_keyvault_reference
                with patch.object(secret_provider, "resolve_keyvault_reference_base") as mock_base:
                    vault_url = "https://myvault.vault.azure.net/"
                    mock_base.return_value = (mock_id_instance, vault_url)

                    result = await secret_provider.resolve_keyvault_reference(config)

                    # Verify the specific credential was used instead of the default
                    mock_client_class.assert_called_once_with(
                        vault_url=vault_url, credential=mock_specific_credential, retry_total=3
                    )
                    # Verify the result
                    self.assertEqual(result, "secret-value")

    @app_config_decorator_async
    @recorded_by_proxy_async
    async def test_integration_with_keyvault(
        self, appconfiguration_endpoint_string, appconfiguration_keyvault_secret_url
    ):
        """Test integration with Key Vault."""
        if not appconfiguration_keyvault_secret_url:
            self.skipTest("No Key Vault secret URL provided")

        # Get a credential
        credential = self.get_credential(SecretClient, is_async=True)

        # Create a SecretProvider with the credential
        secret_provider = SecretProvider(keyvault_credential=credential)

        # Create a Key Vault reference
        config = SecretReferenceConfigurationSetting(key="test-key", secret_id=appconfiguration_keyvault_secret_url)

        # Resolve the reference
        secret_value = await secret_provider.resolve_keyvault_reference(config)

        # Verify a value was returned (we can't know the exact value)
        self.assertIsNotNone(secret_value)
        self.assertTrue(isinstance(secret_value, str))

        # Verify the secret was cached
        self.assertIn(appconfiguration_keyvault_secret_url, secret_provider._secret_cache)


if __name__ == "__main__":
    unittest.main()



# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------

from ._version import VERSION

USER_AGENT = "python-appconfiguration-provider/{}".format(VERSION)



# coding: utf-8
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from devtools_testutils import AzureRecordedTestCase
from test_constants import FEATURE_MANAGEMENT_KEY, FEATURE_FLAG_KEY
from azure.appconfiguration import (
    AzureAppConfigurationClient,
    ConfigurationSetting,
    FeatureFlagConfigurationSetting,
    SecretReferenceConfigurationSetting,
)
from azure.appconfiguration.provider import load, AzureAppConfigurationKeyVaultOptions
from azure.appconfiguration.provider._constants import NULL_CHAR


class AppConfigTestCase(AzureRecordedTestCase):
    def create_client(self, **kwargs):
        credential = self.get_credential(AzureAppConfigurationClient)

        client = None

        if "connection_string" in kwargs:
            client = AzureAppConfigurationClient.from_connection_string(kwargs["connection_string"])
        else:
            client = AzureAppConfigurationClient(kwargs["endpoint"], credential)

        setup_configs(client, kwargs.get("keyvault_secret_url"), kwargs.get("keyvault_secret_url2"))
        kwargs["user_agent"] = "SDK/Integration"

        if "endpoint" in kwargs:
            kwargs["credential"] = credential

        if "secret_resolver" not in kwargs and kwargs.get("keyvault_secret_url") and "key_vault_options" not in kwargs:
            kwargs["keyvault_credential"] = credential

        if "key_vault_options" in kwargs:
            key_vault_options = kwargs.pop("key_vault_options")
            if not key_vault_options.secret_resolver:
                key_vault_options = AzureAppConfigurationKeyVaultOptions(credential=credential)
            kwargs["key_vault_options"] = key_vault_options

        return load(**kwargs)

    @staticmethod
    def create_sdk_client(appconfiguration_connection_string):
        return AzureAppConfigurationClient.from_connection_string(
            appconfiguration_connection_string, user_agent="SDK/Integration"
        )

    def create_aad_sdk_client(self, appconfiguration_endpoint_string):
        cred = self.get_credential(AzureAppConfigurationClient)
        return AzureAppConfigurationClient(appconfiguration_endpoint_string, cred, user_agent="SDK/Integration")


def setup_configs(client, keyvault_secret_url, keyvault_secret_url2):
    for config in get_configs(keyvault_secret_url, keyvault_secret_url2):
        client.set_configuration_setting(config)


def get_configs(keyvault_secret_url, keyvault_secret_url2):
    configs = []
    configs.append(create_config_setting("message", NULL_CHAR, "hi"))
    configs.append(create_config_setting("message", "dev", "test"))
    configs.append(create_config_setting("my_json", NULL_CHAR, '{"key": "value"}', "application/json"))
    configs.append(create_config_setting("test.trimmed", NULL_CHAR, "key"))
    configs.append(create_config_setting("refresh_message", NULL_CHAR, "original value"))
    configs.append(create_config_setting("non_refreshed_message", NULL_CHAR, "Static"))
    configs.append(create_config_setting("tagged_config", NULL_CHAR, None, tags={"a": "b"}))
    configs.append(create_config_setting("two_tagged", NULL_CHAR, None, tags={"a": "b", "second": "tag"}))
    configs.append(create_config_setting("only_second_tag", NULL_CHAR, None, tags={"second": "tag"}))
    configs.append(
        create_config_setting(
            "complex_tag", NULL_CHAR, None, tags={"Special:Tag": "Value:With:Colons", "Tag@With@At": "Value@With@At"}
        )
    )
    configs.append(
        create_config_setting(
            ".appconfig.featureflag/Alpha",
            NULL_CHAR,
            '{	"id": "Alpha", "description": "", "enabled": false, "conditions": {	"client_filters": []	}}',
            "application/vnd.microsoft.appconfig.ff+json;charset=utf-8",
        )
    )
    configs.append(
        create_config_setting(
            ".appconfig.featureflag/TaggedFeatureFlag",
            NULL_CHAR,
            '{	"id": "TaggedFeatureFlag", "description": "", "enabled": false, "conditions": {	"client_filters": []	}}',
            "application/vnd.microsoft.appconfig.ff+json;charset=utf-8",
            tags={"a": "b"},
        )
    )
    # Configuration with multiple tags
    configs.append(create_config_setting("multi_tagged", NULL_CHAR, "multi tagged value", tags={"a": "b", "c": "d"}))
    # Configuration with tag that has special characters
    configs.append(create_config_setting("special_chars", NULL_CHAR, "special", tags={"special@tag": "special:value"}))
    # Configuration with no tags
    configs.append(create_config_setting("no_tags", NULL_CHAR, "no tags"))
    # Configuration with tag that has no value
    configs.append(create_config_setting("tag_no_value", NULL_CHAR, "no value", tags={"a": ""}))
    # Configuration with different tag
    configs.append(create_config_setting("different_tag", NULL_CHAR, "different", tags={"different": "tag"}))
    # Configuration with null tag
    configs.append(create_config_setting("null_tag", NULL_CHAR, "null tag", tags={"tag": None}))
    if keyvault_secret_url:
        configs.append(
            create_secret_config_setting(
                "secret",
                "prod",
                keyvault_secret_url,
            )
        )
    if keyvault_secret_url2:
        configs.append(
            create_secret_config_setting(
                "secret2",
                "prod",
                keyvault_secret_url2,
            )
        )
    return configs


def create_config_setting(key, label, value, content_type="text/plain", tags=None):
    return ConfigurationSetting(key=key, label=label, value=value, content_type=content_type, tags=tags)


def create_secret_config_setting(key, label, value):
    return SecretReferenceConfigurationSetting(
        key=key,
        label=label,
        secret_id=value,
    )


def create_feature_flag_config_setting(key, label, enabled, tags=None):
    return FeatureFlagConfigurationSetting(feature_id=key, label=label, enabled=enabled, tags=tags)


def cleanup_test_resources(
    client,
    settings=None,
    snapshot_names=None,
):
    """
    Clean up test resources from Azure App Configuration.

    :param client: The AzureAppConfigurationClient to use for cleanup.
    :param settings: List of ConfigurationSetting objects to delete.
    :param snapshot_names: List of snapshot names to archive.
    """
    # Archive snapshots
    if snapshot_names:
        for snapshot_name in snapshot_names:
            try:
                client.archive_snapshot(snapshot_name)
            except Exception:
                pass

    # Delete configuration settings and feature flags
    if settings:
        for setting in settings:
            try:
                client.delete_configuration_setting(key=setting.key, label=setting.label)
            except Exception:
                pass


def set_test_settings(client, settings):
    """
    Set multiple configuration settings in Azure App Configuration.

    :param client: The AzureAppConfigurationClient to use.
    :param settings: List of ConfigurationSetting or FeatureFlagConfigurationSetting objects to set.
    """
    for setting in settings:
        client.set_configuration_setting(setting)


def create_snapshot(client, snapshot_name, key_filters, composition_type=None, retention_period=3600):
    """
    Create a snapshot in Azure App Configuration and verify it was created successfully.

    :param client: The AzureAppConfigurationClient to use.
    :param snapshot_name: The name of the snapshot to create.
    :param key_filters: List of key filter strings to define what settings to include.
    :param composition_type: The composition type for the snapshot (default: SnapshotComposition.KEY).
    :param retention_period: The retention period in seconds (default: 3600, minimum valid value).
    :return: The created snapshot.
    """
    from azure.appconfiguration import SnapshotComposition, ConfigurationSettingsFilter, SnapshotStatus
    from devtools_testutils import is_live

    if composition_type is None:
        composition_type = SnapshotComposition.KEY

    # Convert key filter strings to ConfigurationSettingsFilter objects
    filters = [ConfigurationSettingsFilter(key=key_filter) for key_filter in key_filters]

    snapshot = client.begin_create_snapshot(
        name=snapshot_name,
        filters=filters,
        composition_type=composition_type,
        retention_period=retention_period,
    ).result()

    # Verify snapshot was created successfully
    assert snapshot.status == SnapshotStatus.READY, f"Snapshot status is {snapshot.status}, expected READY"
    assert (
        snapshot.composition_type == composition_type
    ), f"Snapshot composition_type is {snapshot.composition_type}, expected {composition_type}"

    # Verify snapshot name (sanitized in playback mode)
    if is_live():
        assert snapshot.name == snapshot_name, f"Snapshot name is {snapshot.name}, expected {snapshot_name}"
    else:
        assert snapshot.name == "Sanitized", f"Snapshot name is {snapshot.name}, expected 'Sanitized' in playback"

    return snapshot


def get_feature_flag(client, feature_id):
    for feature_flag in client[FEATURE_MANAGEMENT_KEY][FEATURE_FLAG_KEY]:
        if feature_flag["id"] == feature_id:
            return feature_flag
    return None


def has_feature_flag(client, feature_id, enabled=False):
    feature_flag = get_feature_flag(client, feature_id)
    if feature_flag:
        return feature_flag["enabled"] == enabled
    return False



# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------

from ._secret_provider import SecretProvider
from ._secret_provider_base import _SecretProviderBase

__all__ = [
    "SecretProvider",
    "_SecretProviderBase",
]



# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import os
from sample_utilities import get_client_modifications
from azure.appconfiguration.provider import load, SettingSelector

kwargs = get_client_modifications()
connection_string = os.environ["APPCONFIGURATION_CONNECTION_STRING"]

# Connecting to Azure App Configuration using connection string
config = load(connection_string=connection_string, **kwargs)

print(config["message"])
print(config["my_json"]["key"])

# Connecting to Azure App Configuration using connection string and trimmed key prefixes
trimmed = ["test."]
config = load(connection_string=connection_string, trim_prefixes=trimmed, **kwargs)

print(config["message"])

# Connection to Azure App Configuration using SettingSelector
selects = [SettingSelector(key_filter="message*")]
config = load(connection_string=connection_string, selects=selects, **kwargs)

print("message found: " + str("message" in config))
print("test.message found: " + str("test.message" in config))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
from unittest.mock import Mock, patch, MagicMock
from azure.appconfiguration import ConfigurationSetting, SnapshotComposition
from azure.appconfiguration.provider._snapshot_reference_parser import SnapshotReferenceParser
from azure.appconfiguration.provider._constants import (
    SNAPSHOT_REF_CONTENT_TYPE,
)
from azure.appconfiguration.provider._client_manager import _ConfigurationClientWrapper
from azure.core.exceptions import HttpResponseError


class TestSnapshotReferenceParser:
    """Tests for the SnapshotReferenceParser class."""

    def test_parse_valid_snapshot_reference(self):
        """Test parsing a valid snapshot reference."""
        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value='{"snapshot_name": "test-snapshot"}',
            content_type=SNAPSHOT_REF_CONTENT_TYPE,
        )

        result = SnapshotReferenceParser.parse(setting)

        assert isinstance(result, str)
        assert result == "test-snapshot"

    def test_parse_snapshot_reference_with_whitespace(self):
        """Test parsing a snapshot reference with whitespace in snapshot name."""
        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value='{"snapshot_name": "  test-snapshot  "}',
            content_type=SNAPSHOT_REF_CONTENT_TYPE,
        )

        result = SnapshotReferenceParser.parse(setting)

        assert isinstance(result, str)
        assert result == "test-snapshot"  # Should be trimmed

    def test_parse_none_setting_raises_error(self):
        """Test that parsing None setting raises ValueError."""
        with pytest.raises(ValueError, match="Setting cannot be None"):
            SnapshotReferenceParser.parse(None)

    def test_parse_empty_value_raises_error(self):
        """Test that parsing empty value raises ValueError."""
        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value="{}",
            content_type=SNAPSHOT_REF_CONTENT_TYPE,
        )

        with pytest.raises(ValueError, match="Invalid snapshot reference format.*snapshot_name.*is required"):
            SnapshotReferenceParser.parse(setting)

    def test_parse_whitespace_value_raises_error(self):
        """Test that parsing whitespace-only value raises ValueError."""
        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value="   ",
            content_type=SNAPSHOT_REF_CONTENT_TYPE,
        )

        with pytest.raises(ValueError, match="Invalid snapshot reference format.*Value cannot be empty"):
            SnapshotReferenceParser.parse(setting)

    def test_parse_invalid_json_raises_error(self):
        """Test that parsing invalid JSON raises ValueError."""
        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value="{invalid json, missing quotes}",
            content_type=SNAPSHOT_REF_CONTENT_TYPE,
        )

        with pytest.raises(ValueError, match="Invalid snapshot reference format.*Invalid JSON format"):
            SnapshotReferenceParser.parse(setting)

    def test_parse_non_object_json_raises_error(self):
        """Test that parsing non-object JSON raises ValueError."""
        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value='"just a string"',
            content_type=SNAPSHOT_REF_CONTENT_TYPE,
        )

        with pytest.raises(ValueError, match="Invalid snapshot reference format.*Expected JSON object"):
            SnapshotReferenceParser.parse(setting)

    def test_parse_missing_snapshot_name_property_raises_error(self):
        """Test that missing snapshot_name property raises ValueError."""
        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value='{"other_property": "value"}',
            content_type=SNAPSHOT_REF_CONTENT_TYPE,
        )

        with pytest.raises(ValueError, match="Invalid snapshot reference format.*property is required"):
            SnapshotReferenceParser.parse(setting)

    def test_parse_non_string_snapshot_name_raises_error(self):
        """Test that non-string snapshot_name property raises ValueError."""
        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value='{"snapshot_name": 123}',
            content_type=SNAPSHOT_REF_CONTENT_TYPE,
        )

        with pytest.raises(ValueError, match="Invalid snapshot reference format.*must be a string value"):
            SnapshotReferenceParser.parse(setting)

    def test_parse_empty_snapshot_name_raises_error(self):
        """Test that empty snapshot_name raises ValueError."""
        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value='{"snapshot_name": ""}',
            content_type=SNAPSHOT_REF_CONTENT_TYPE,
        )

        with pytest.raises(ValueError, match="Invalid snapshot reference format.*cannot be empty or whitespace"):
            SnapshotReferenceParser.parse(setting)

    def test_parse_whitespace_snapshot_name_raises_error(self):
        """Test that whitespace-only snapshot_name raises ValueError."""
        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value='{"snapshot_name": "   "}',
            content_type=SNAPSHOT_REF_CONTENT_TYPE,
        )

        with pytest.raises(ValueError, match="Invalid snapshot reference format.*cannot be empty or whitespace"):
            SnapshotReferenceParser.parse(setting)


class TestResolveSnapshotReference:
    """Tests for the resolve_snapshot_reference method in _ConfigurationClientWrapper."""

    def test_resolve_snapshot_reference_success(self):
        """Test successfully resolving a snapshot reference."""
        # Create mock client
        mock_app_config_client = Mock()
        wrapper = _ConfigurationClientWrapper("http://test.endpoint", mock_app_config_client)

        # Mock snapshot
        mock_snapshot = Mock()
        mock_snapshot.composition_type = SnapshotComposition.KEY
        mock_app_config_client.get_snapshot.return_value = mock_snapshot

        # Mock configuration settings from snapshot
        mock_settings = [
            ConfigurationSetting(key="key1", value="value1"),
            ConfigurationSetting(key="key2", value="value2"),
        ]

        with patch.object(wrapper, "load_configuration_settings", return_value=mock_settings):
            setting = ConfigurationSetting(
                key="SnapshotRef1",
                value='{"snapshot_name": "test-snapshot"}',
                content_type=SNAPSHOT_REF_CONTENT_TYPE,
            )

            result = wrapper.resolve_snapshot_reference(setting)

            assert result == mock_settings
            mock_app_config_client.get_snapshot.assert_called_once_with("test-snapshot")

    def test_resolve_snapshot_reference_not_found_returns_empty_list(self):
        """Test that 404 error when resolving snapshot reference returns empty list."""
        # Create mock client that raises 404
        mock_app_config_client = Mock()
        wrapper = _ConfigurationClientWrapper("http://test.endpoint", mock_app_config_client)

        # Mock get_snapshot to raise 404
        http_error = HttpResponseError("Not Found")
        http_error.status_code = 404
        mock_app_config_client.get_snapshot.side_effect = http_error

        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value='{"snapshot_name": "non-existent-snapshot"}',
            content_type=SNAPSHOT_REF_CONTENT_TYPE,
        )

        result = wrapper.resolve_snapshot_reference(setting)

        assert result == []
        mock_app_config_client.get_snapshot.assert_called_once_with("non-existent-snapshot")

    def test_resolve_snapshot_reference_non_404_error_raises(self):
        """Test that non-404 HTTP errors are raised."""
        # Create mock client that raises 500
        mock_app_config_client = Mock()
        wrapper = _ConfigurationClientWrapper("http://test.endpoint", mock_app_config_client)

        # Mock get_snapshot to raise 500 error
        http_error = HttpResponseError("Internal Server Error")
        http_error.status_code = 500
        mock_app_config_client.get_snapshot.side_effect = http_error

        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value='{"snapshot_name": "test-snapshot"}',
            content_type=SNAPSHOT_REF_CONTENT_TYPE,
        )

        with pytest.raises(HttpResponseError) as exc_info:
            wrapper.resolve_snapshot_reference(setting)

        assert exc_info.value.status_code == 500

    def test_resolve_snapshot_reference_invalid_composition_type_raises(self):
        """Test that invalid snapshot composition type raises ValueError."""
        # Create mock client
        mock_app_config_client = Mock()
        wrapper = _ConfigurationClientWrapper("http://test.endpoint", mock_app_config_client)

        # Mock snapshot with invalid composition type
        mock_snapshot = Mock()
        mock_snapshot.composition_type = "invalid"  # Not SnapshotComposition.KEY
        mock_app_config_client.get_snapshot.return_value = mock_snapshot

        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value='{"snapshot_name": "test-snapshot"}',
            content_type=SNAPSHOT_REF_CONTENT_TYPE,
        )

        with pytest.raises(ValueError, match="Composition type for 'test-snapshot' must be 'key'."):
            wrapper.resolve_snapshot_reference(setting)

    def test_resolve_snapshot_reference_invalid_content_type_raises(self):
        """Test that invalid content type raises ValueError."""
        # Create mock client
        mock_app_config_client = Mock()
        wrapper = _ConfigurationClientWrapper("http://test.endpoint", mock_app_config_client)

        setting = ConfigurationSetting(
            key="SnapshotRef1",
            value='{"snapshot_name": "test-snapshot"}',
            content_type="application/json",  # Wrong content type
        )

        with pytest.raises(ValueError, match="Setting is not a snapshot reference"):
            wrapper.resolve_snapshot_reference(setting)
